from json import dumps
from .legacyencoder import legacy_dict_from_doc


def encode(document, helixOrderList, io):
    obj = legacy_dict_from_doc(document, io.name, helixOrderList)
    json_string = dumps(obj, separators=(',',':'))  # compact encoding
    io.write(json_string)



#!/usr/bin/env python
# encoding: utf-8

# from exceptions import NotImplementedError
from math import floor
from cadnano2.views import styles

import cadnano2.views.pathview.pathselection as pathselection

import cadnano2.util as util
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'QObject', 'QPointF',
                                        'QRectF', 'Qt'])
util.qtWrapImport('QtGui', globals(), ['QBrush', 'QColor', 'QPen',
                                       'QPainterPath', 'QPolygonF'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsPathItem',
                                           'QGraphicsItem',
                                           'QGraphicsRectItem'])

_baseWidth = styles.PATH_BASE_WIDTH

ppL5 = QPainterPath()  # Left 5' PainterPath
ppR5 = QPainterPath()  # Right 5' PainterPath
ppL3 = QPainterPath()  # Left 3' PainterPath
ppR3 = QPainterPath()  # Right 3' PainterPath
pp53 = QPainterPath()  # Left 5', Right 3' PainterPath
pp35 = QPainterPath()  # Left 5', Right 3' PainterPath
# set up ppL5 (left 5' blue square)
ppL5.addRect(0.25 * _baseWidth,
             0.125 * _baseWidth,
             0.75 * _baseWidth,
             0.75 * _baseWidth)
# set up ppR5 (right 5' blue square)
ppR5.addRect(0, 0.125 * _baseWidth, 0.75 * _baseWidth, 0.75 * _baseWidth)
# set up ppL3 (left 3' blue triangle)
l3poly = QPolygonF()
l3poly.append(QPointF(_baseWidth, 0))
l3poly.append(QPointF(0.25 * _baseWidth, 0.5 * _baseWidth))
l3poly.append(QPointF(_baseWidth, _baseWidth))
l3poly.append(QPointF(_baseWidth, 0))
ppL3.addPolygon(l3poly)
# set up ppR3 (right 3' blue triangle)
r3poly = QPolygonF()
r3poly.append(QPointF(0, 0))
r3poly.append(QPointF(0.75 * _baseWidth, 0.5 * _baseWidth))
r3poly.append(QPointF(0, _baseWidth))
r3poly.append(QPointF(0, 0))
ppR3.addPolygon(r3poly)

# single base left 5'->3'
pp53.addRect(0, 0.125 * _baseWidth, 0.5 * _baseWidth, 0.75 * _baseWidth)
poly53 = QPolygonF()
poly53.append(QPointF(0.5 * _baseWidth, 0))
poly53.append(QPointF(_baseWidth, 0.5 * _baseWidth))
poly53.append(QPointF(0.5 * _baseWidth, _baseWidth))
pp53.addPolygon(poly53)
# single base left 3'<-5'
pp35.addRect(0.50 * _baseWidth,
             0.125 * _baseWidth,
             0.5 * _baseWidth,
             0.75 * _baseWidth)
poly35 = QPolygonF()
poly35.append(QPointF(0.5 * _baseWidth, 0))
poly35.append(QPointF(0, 0.5 * _baseWidth))
poly35.append(QPointF(0.5 * _baseWidth, _baseWidth))
pp35.addPolygon(poly35)

_defaultRect = QRectF(0, 0, _baseWidth, _baseWidth)
_noPen = QPen(Qt.PenStyle.NoPen)


class EndpointItem(QGraphicsPathItem):

    _filterName = "endpoint"

    def __init__(self, strandItem, captype, isDrawn5to3):
        """The parent should be a StrandItem."""
        super(EndpointItem, self).__init__(strandItem.virtualHelixItem())

        self._strandItem = strandItem
        self._activeTool = strandItem.activeTool()
        self._capType = captype
        self._lowDragBound = None
        self._highDragBound = None
        self._initCapSpecificState(isDrawn5to3)
        self.setPen(QPen())
        # for easier mouseclick
        self._clickArea = cA = QGraphicsRectItem(_defaultRect, self)
        self._clickArea.setAcceptHoverEvents(True)
        cA.hoverMoveEvent = self.hoverMoveEvent
        cA.mousePressEvent = self.mousePressEvent
        cA.mouseMoveEvent = self.mouseMoveEvent
        cA.setPen(_noPen)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
    # end def

    # def __repr__(self):
    #     return "%s" % self.__class__.__name__

    ### SIGNALS ###

    ### SLOTS ###

    ### ACCESSORS ###
    def idx(self):
        """Look up baseIdx, as determined by strandItem idxs and cap type."""
        if self._capType == 'low':
            return self._strandItem.idxs()[0]
        else:  # high or dual, doesn't matter
            return self._strandItem.idxs()[1]
    # end def

    def partItem(self):
        return self._strandItem.partItem()
    # end def

    def disableEvents(self):
        self._clickArea.setAcceptHoverEvents(False)
        self.mouseMoveEvent = QGraphicsPathItem.mouseMoveEvent
        self.mousePressEvent = QGraphicsPathItem.mousePressEvent
    # end def

    def window(self):
        return self._strandItem.window()

    ### PUBLIC METHODS FOR DRAWING / LAYOUT ###
    def updatePosIfNecessary(self, idx):
        """Update position if necessary and return True if updated."""
        group = self.group()
        self.tempReparent()
        x = int(idx * _baseWidth)
        if x != self.x():
            self.setPos(x, self.y())
            if group:
                group.addToGroup(self)
            return True
        else:
            if group:
                group.addToGroup(self)
            return False

    def safeSetPos(self, x, y):
        """
        Required to ensure proper reparenting if selected
        """
        group = self.group()
        self.tempReparent()
        self.setPos(x,y)
        if group:
            group.addToGroup(self)
    # end def

    def resetEndPoint(self, isDrawn5to3):
        self.setParentItem(self._strandItem.virtualHelixItem())
        self._initCapSpecificState(isDrawn5to3)
        upperLeftY = 0 if isDrawn5to3 else _baseWidth
        self.setY(upperLeftY)
    # end def

    ### PRIVATE SUPPORT METHODS ###
    def _initCapSpecificState(self, isDrawn5to3):
        cT = self._capType
        if cT == 'low':
            path = ppL5 if isDrawn5to3 else ppL3
        elif cT == 'high':
            path = ppR3 if isDrawn5to3 else ppR5
        elif cT == 'dual':
            path = pp53 if isDrawn5to3 else pp35
        self.setPath(path)
    # end def

    def _getNewIdxsForResize(self, baseIdx):
        """Returns a tuple containing idxs to be passed to the """
        cT = self._capType
        if cT == 'low':
            return (baseIdx, self._strandItem.idxs()[1])
        elif cT == 'high':
            return (self._strandItem.idxs()[0], baseIdx)
        elif cT == 'dual':
            raise NotImplementedError

    ### EVENT HANDLERS ###
    def mousePressEvent(self, event):
        """
        Parses a mousePressEvent, calling the approproate tool method as
        necessary. Stores _moveIdx for future comparison.
        """
        self.scene().views()[0].addToPressList(self)
        self._strandItem.virtualHelixItem().setActive(self.idx())
        self._moveIdx = self.idx()
        activeToolStr = str(self._activeTool())
        toolMethodName = activeToolStr + "MousePress"
        if hasattr(self, toolMethodName):  # if the tool method exists
            modifiers = event.modifiers()
            getattr(self, toolMethodName)(modifiers, event, self.idx())

    def hoverLeaveEvent(self, event):
        self.partItem().updateStatusBar("")
    # end def

    def hoverMoveEvent(self, event):
        """
        Parses a mousePressEvent, calling the approproate tool method as
        necessary. Stores _moveIdx for future comparison.
        """
        vhiNum = self._strandItem._virtualHelixItem.number()
        oligoLength = self._strandItem._modelStrand.oligo().length()
        msg = "%d[%d]\tlength: %d" % (vhiNum, self.idx(), oligoLength)
        self.partItem().updateStatusBar(msg)

        activeToolStr = str(self._activeTool())
        if activeToolStr == 'pencilTool':
            return self._strandItem.pencilToolHoverMove(self.idx())

    def mouseMoveEvent(self, event):
        """
        Parses a mouseMoveEvent, calling the approproate tool method as
        necessary. Updates _moveIdx if it changed.
        """
        toolMethodName = str(self._activeTool()) + "MouseMove"
        if hasattr(self, toolMethodName):  # if the tool method exists
            idx = int(floor((self.x() + event.pos().x()) / _baseWidth))
            if idx != self._moveIdx:  # did we actually move?
                modifiers = event.modifiers()
                self._moveIdx = idx
                getattr(self, toolMethodName)(modifiers, idx)

    def customMouseRelease(self, event):
        """
        Parses a mouseReleaseEvent, calling the approproate tool method as
        necessary. Deletes _moveIdx if necessary.
        """
        toolMethodName = str(self._activeTool()) + "MouseRelease"
        if hasattr(self, toolMethodName):  # if the tool method exists
            modifiers = event.modifiers()
            x = event.position().x()
            getattr(self, toolMethodName)(modifiers, x)  # call tool method
        if self._moveIdx is not None:
            self._moveIdx = None

    ### TOOL METHODS ###
    def addSeqToolMousePress(self, modifiers, event, idx):
        """
        Checks that a scaffold was clicked, and then calls apply sequence
        to the clicked strand via its oligo.
        """
        mStrand = self._strandItem._modelStrand
        if mStrand.isScaffold():
            olgLen, seqLen = self._activeTool().applySequence(mStrand.oligo())
            if olgLen:
                msg = "Populated %d of %d scaffold bases." % (min(seqLen, olgLen), olgLen)
                if olgLen > seqLen:
                    d = olgLen - seqLen
                    msg = msg + " Warning: %d bases have no sequence." % d
                elif olgLen < seqLen:
                    d = seqLen - olgLen
                    msg = msg + " Warning: %d sequence bases unused." % d
                self.partItem().updateStatusBar(msg)


    def breakToolMouseRelease(self, modifiers, x):
        """Shift-click to merge without switching back to select tool."""
        mStrand = self._strandItem._modelStrand
        if modifiers & Qt.KeyboardModifier.ShiftModifier:
            mStrand.merge(self.idx())
    # end def

    def eraseToolMousePress(self, modifiers, event, idx):
        """Erase the strand."""
        mStrand = self._strandItem._modelStrand
        mStrand.strandSet().removeStrand(mStrand)
    # end def

    def insertionToolMousePress(self, modifiers, event, idx):
        """Add an insert to the strand if possible."""
        mStrand = self._strandItem._modelStrand
        mStrand.addInsertion(idx, 1)
    # end def

    def paintToolMousePress(self, modifiers, event, idx):
        """Add an insert to the strand if possible."""
        mStrand = self._strandItem._modelStrand
        if mStrand.isStaple():
            color = self.window().pathColorPanel.stapColorName()
        else:
            color = self.window().pathColorPanel.scafColorName()
        mStrand.oligo().applyColor(color)
    # end def

    def pencilToolHoverMove(self, idx):
        """Pencil the strand is possible."""
        mStrand = self._strandItem._modelStrand
        vhi = self._strandItem._virtualHelixItem
        activeTool = self._activeTool()

        if not activeTool.isFloatingXoverBegin():
            tempXover = activeTool.floatingXover()
            tempXover.updateFloatingFromStrandItem(vhi, mStrand, idx)
            # if mStrand.idx5Prime() == idx:
            #     tempXover.hide3prime()
    # end def

    def pencilToolMousePress(self, modifiers, event, idx):
        """Break the strand is possible."""
        mStrand = self._strandItem._modelStrand
        vhi = self._strandItem._virtualHelixItem
        partItem = vhi.partItem()
        activeTool = self._activeTool()

        if activeTool.isFloatingXoverBegin():
            if mStrand.idx5Prime() == idx:
                return
            tempXover = activeTool.floatingXover()
            tempXover.updateBase(vhi, mStrand, idx)
            activeTool.setFloatingXoverBegin(False)
            # tempXover.hide5prime()
        else:
            activeTool.setFloatingXoverBegin(True)
            # install Xover
            activeTool.attemptToCreateXover(vhi, mStrand, idx)
    # end def

    # def selectToolMousePress(self, modifiers, event):
    #     """
    #     Set the allowed drag bounds for use by selectToolMouseMove.
    #     """
    #     print "mouse press ep", self.parentItem()
    #     # print "%s.%s [%d]" % (self, util.methodName(), self.idx())
    #     self._lowDragBound, self._highDragBound = \
    #                 self._strandItem._modelStrand.getResizeBounds(self.idx())
    #     sI = self._strandItem
    #     viewroot = sI.viewroot()
    #     selectionGroup = viewroot.strandItemSelectionGroup()
    #     selectionGroup.setInstantAdd(True)
    #     self.setSelected(True)
    # # end def

    def selectToolMousePress(self, modifiers, event, idx):
        """
        Set the allowed drag bounds for use by selectToolMouseMove.
        """
        # print "%s.%s [%d]" % (self, util.methodName(), self.idx())
        self._lowDragBound, self._highDragBound = \
                    self._strandItem._modelStrand.getResizeBounds(self.idx())
        sI = self._strandItem
        viewroot = sI.viewroot()
        currentFilterDict = viewroot.selectionFilterDict()
        if sI.strandFilter() in currentFilterDict \
                                    and self._filterName in currentFilterDict:
            selectionGroup = viewroot.strandItemSelectionGroup()
            mod = Qt.KeyboardModifier.MetaModifier
            if not (modifiers & mod):
                selectionGroup.clearSelection(False)
            selectionGroup.setSelectionLock(selectionGroup)
            selectionGroup.pendToAdd(self)
            selectionGroup.processPendingToAddList()
            return selectionGroup.mousePressEvent(event)
    # end def

    def selectToolMouseMove(self, modifiers, idx):
        """
        Given a new index (pre-validated as different from the prev index),
        calculate the new x coordinate for self, move there, and notify the
        parent strandItem to redraw its horizontal line.
        """
        idx = util.clamp(idx, self._lowDragBound, self._highDragBound)
        # x = int(idx * _baseWidth)
        # self.setPos(x, self.y())
        # self._strandItem.updateLine(self)
    # end def

    def selectToolMouseRelease(self, modifiers, x):
        """
        If the positional-calculated idx differs from the model idx, it means
        we have moved and should notify the model to resize.

        If the mouse event had a key modifier, perform special actions:
            shift = attempt to merge with a neighbor
            alt = extend to max drag bound
        """
        mStrand = self._strandItem._modelStrand
        baseIdx = int(floor(self.x() / _baseWidth))
        # if baseIdx != self.idx():
        #     newIdxs = self._getNewIdxsForResize(baseIdx)
        #     mStrand.resize(newIdxs)

        if modifiers & Qt.KeyboardModifier.AltModifier:
            if self._capType == 'low':
                newIdxs = self._getNewIdxsForResize(self._lowDragBound)
            else:
                newIdxs = self._getNewIdxsForResize(self._highDragBound)
            mStrand.resize(newIdxs)
        elif modifiers & Qt.KeyboardModifier.ShiftModifier:
            self.setSelected(False)
            self.restoreParent()
            mStrand.merge(self.idx())
    # end def

    def skipToolMousePress(self, modifiers, event, idx):
        """Add an insert to the strand if possible."""
        mStrand = self._strandItem._modelStrand
        mStrand.addInsertion(idx, -1)
    # end def

    def restoreParent(self, pos=None):
        """
        Required to restore parenting and positioning in the partItem
        """
        # map the position
        # print "restoring parent ep"
        self.tempReparent(pos)
        self.setSelectedColor(False)
        self.setSelected(False)
    # end def

    def tempReparent(self, pos=None):
        vhItem = self._strandItem.virtualHelixItem()
        if pos == None:
            pos = self.scenePos()
        self.setParentItem(vhItem)
        tempP = vhItem.mapFromScene(pos)
        self.setPos(tempP)
    # end def

    def setSelectedColor(self, value):
        if value == True:
            color = styles.selected_color
        else:
            oligo = self._strandItem.strand().oligo()
            color = QColor(oligo.color())
            if oligo.shouldHighlight():
                color.setAlpha(128)
        brush = self.brush()
        brush.setColor(color)
        self.setBrush(brush)
    # end def

    def updateHighlight(self, brush):
        if not self.isSelected():
            self.setBrush(brush)
    # end def

    def itemChange(self, change, value):
        # for selection changes test against QGraphicsItem.GraphicsItemChange.ItemSelectedChange
        # intercept the change instead of the has changed to enable features.
        if change == QGraphicsItem.GraphicsItemChange.ItemSelectedChange and self.scene():
            activeTool = self._activeTool()
            if str(activeTool) == "selectTool":
                sI = self._strandItem
                viewroot = sI.viewroot()
                currentFilterDict = viewroot.selectionFilterDict()
                selectionGroup = viewroot.strandItemSelectionGroup()

                # only add if the selectionGroup is not locked out
                if value == True and self._filterName in currentFilterDict:
                    # if self.group() != selectionGroup \
                    #                   and sI.strandFilter() in currentFilterDict:
                    if sI.strandFilter() in currentFilterDict:
                        if self.group() != selectionGroup or not self.isSelected():
                            selectionGroup.pendToAdd(self)
                            selectionGroup.setSelectionLock(selectionGroup)
                            self.setSelectedColor(True)
                        return True
                    else:
                        return False
                # end if
                elif value == True:
                    # don't select
                    return False
                else:
                    # Deselect
                    # Check if strand is being added to the selection group still
                    if not selectionGroup.isPending(self._strandItem):
                        selectionGroup.pendToRemove(self)
                        self.tempReparent()
                        self.setSelectedColor(False)
                        return False
                    else:   # don't deselect, because the strand is still selected
                        return True
                # end else
            # end if
            elif str(activeTool) == "paintTool":
                sI = self._strandItem
                viewroot = sI.viewroot()
                currentFilterDict = viewroot.selectionFilterDict()
                if sI.strandFilter() in currentFilterDict:
                    if not activeTool.isMacrod():
                        activeTool.setMacrod()
                    self.paintToolMousePress(None, None, None)
            # end elif
            return False
        # end if
        return QGraphicsPathItem.itemChange(self, change, value)
    # end def

    def modelDeselect(self, document):
        strand = self._strandItem.strand()
        test = document.isModelStrandSelected(strand)
        lowVal, highVal = document.getSelectedStrandValue(strand) if test \
                                                            else (False, False)
        if self._capType == 'low':
            outValue = (False, highVal)
        else:
            outValue = (lowVal, False)
        if not outValue[0] and not outValue[1] and test:
            document.removeStrandFromSelection(strand)
        elif outValue[0] or outValue[1]:
            document.addStrandToSelection(strand, outValue)
        self.restoreParent()
    # end def

    def modelSelect(self, document):
        strand = self._strandItem.strand()
        test = document.isModelStrandSelected(strand)
        lowVal, highVal = document.getSelectedStrandValue(strand) if test \
                                                            else (False, False)
        if self._capType == 'low':
            outValue = (True, highVal)
        else:
            outValue = (lowVal, True)
        self.setSelected(True)
        self.setSelectedColor(True)
        document.addStrandToSelection(strand, outValue)
    # end def

    def paint(self, painter, option, widget):
        painter.setPen(self.pen())
        painter.setBrush(self.brush())
        painter.drawPath(self.path())
    # end def



from cadnano2.cadnano import app
from .pathview.colorpanel import ColorPanel
from .pathview.tools.pathtoolmanager import PathToolManager
from .sliceview.slicerootitem import SliceRootItem
from .pathview.pathrootitem import PathRootItem

from .sliceview.tools.slicetoolmanager import SliceToolManager
import cadnano2.ui.mainwindow.ui_mainwindow as ui_mainwindow
import cadnano2.util as util

util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'Qt', 'QFileInfo',
                                        'QPoint', 'QSettings', 'QSize',
                                        ])
util.qtWrapImport('QtGui', globals(), ['QAction', 'QPaintEngine'])
util.qtWrapImport('QtWidgets', globals(), [
                                           
                                           'QApplication',
                                           'QGraphicsObject',
                                           'QGraphicsScene',
                                           'QGraphicsView',
                                           'QGraphicsItem',
                                           'QGraphicsRectItem',
                                           'QMainWindow',
                                           'QWidget',
                                           ])

# for OpenGL mode
try:
    from OpenGL import GL
except:
    GL = False

GL = False


class DocumentWindow(QMainWindow, ui_mainwindow.Ui_MainWindow):
    """docstring for DocumentWindow"""
    def __init__(self, parent=None, docCtrlr=None):
        super(DocumentWindow, self).__init__(parent)
        self.controller = docCtrlr
        doc = docCtrlr.document()
        self.setupUi(self)
        self.settings = QSettings()
        self._readSettings()
        # Slice setup
        self.slicescene = QGraphicsScene(parent=self.sliceGraphicsView)
        self.sliceroot = SliceRootItem(rect=self.slicescene.sceneRect(),\
                                       parent=None,\
                                       window=self,\
                                       document=doc)
        self.sliceroot.setFlag(QGraphicsItem.GraphicsItemFlag.ItemHasNoContents)
        self.slicescene.addItem(self.sliceroot)
        self.slicescene.setItemIndexMethod(QGraphicsScene.ItemIndexMethod.NoIndex)
        assert self.sliceroot.scene() == self.slicescene
        self.sliceGraphicsView.setScene(self.slicescene)
        self.sliceGraphicsView.sceneRootItem = self.sliceroot
        self.sliceGraphicsView.setName("SliceView")
        self.sliceToolManager = SliceToolManager(self)
        # Path setup
        self.pathscene = QGraphicsScene(parent=self.pathGraphicsView)
        self.pathroot = PathRootItem(rect=self.pathscene.sceneRect(),\
                                     parent=None,\
                                     window=self,\
                                     document=doc)
        self.pathroot.setFlag(QGraphicsItem.GraphicsItemFlag.ItemHasNoContents)
        self.pathscene.addItem(self.pathroot)
        self.pathscene.setItemIndexMethod(QGraphicsScene.ItemIndexMethod.NoIndex)
        assert self.pathroot.scene() == self.pathscene
        self.pathGraphicsView.setScene(self.pathscene)
        self.pathGraphicsView.sceneRootItem = self.pathroot
        self.pathGraphicsView.setScaleFitFactor(0.9)
        self.pathGraphicsView.setName("PathView")
        self.pathColorPanel = ColorPanel()
        self.pathGraphicsView.toolbar = self.pathColorPanel  # HACK for customqgraphicsview
        self.pathscene.addItem(self.pathColorPanel)
        self.pathToolManager = PathToolManager(self)
        self.sliceToolManager.pathToolManager = self.pathToolManager
        self.pathToolManager.sliceToolManager = self.sliceToolManager
        self.tool_managers = (self.sliceToolManager, self.pathToolManager)

        # set the selection filter default
        doc.documentSelectionFilterChangedSignal.emit(["endpoint", "scaffold", "staple", "xover"])

        self.pathGraphicsView.setupGL()
        self.sliceGraphicsView.setupGL()
        if GL:
            pass
            # self.slicescene.drawBackground = self.drawBackgroundGL
            # self.pathscene.drawBackground = self.drawBackgroundGL

        if app().isInMaya():
            from .solidview.solidrootitem import SolidRootItem
            self.splitter.setOrientation(Qt.Vertical)
            self.setUnifiedTitleAndToolBarOnMac(False)
            modState = self.actionModify.isChecked()
            self.solidroot = SolidRootItem(parent=None, document=doc,
                                           modState=modState)

        # Edit menu setup
        self.actionUndo = docCtrlr.undoStack().createUndoAction(self)
        self.actionRedo = docCtrlr.undoStack().createRedoAction(self)
        self.actionUndo.setText(QApplication.translate("MainWindow", "Undo", None))
        self.actionUndo.setShortcut(QApplication.translate("MainWindow", "Ctrl+Z", None))
        self.actionRedo.setText(QApplication.translate("MainWindow", "Redo", None))
        self.actionRedo.setShortcut(QApplication.translate("MainWindow", "Ctrl+Shift+Z", None))

        self.sep = QAction(self)
        self.sep.setSeparator(True)
        self.menuEdit.insertAction(self.actionModify, self.sep)
        self.menuEdit.insertAction(self.sep, self.actionRedo)
        self.menuEdit.insertAction(self.actionRedo, self.actionUndo)
        self.splitter.setSizes([400, 400])  # balance splitter size
        self.statusBar().showMessage("")

    def destroyWin(self):
        self.settings.beginGroup("MainWindow")
        self.settings.setValue("state", self.saveState())
        self.settings.endGroup()
        self.controller = None

    ### ACCESSORS ###
    def undoStack(self):
        return self.controller.undoStack()

    def selectedPart(self):
        return self.controller.document().selectedPart()

    def activateSelection(self, isActive):
        self.pathGraphicsView.activateSelection(isActive)
        self.sliceGraphicsView.activateSelection(isActive)
    # end def

    ### EVENT HANDLERS ###
    def focusInEvent(self):
        app().undoGroup.setActiveStack(self.controller.undoStack())

    def moveEvent(self, event):
        """Reimplemented to save state on move."""
        self.settings.beginGroup("MainWindow")
        self.settings.setValue("pos", self.pos())
        self.settings.endGroup()

    def resizeEvent(self, event):
        """Reimplemented to save state on resize."""
        self.settings.beginGroup("MainWindow")
        self.settings.setValue("size", self.size())
        self.settings.endGroup()
        QWidget.resizeEvent(self, event)

    def changeEvent(self, event):
        QWidget.changeEvent(self, event)
    # end def

    ### DRAWING RELATED ###
    # def drawBackgroundGL(self, painter, rect):
    #     """
    #     This method is for overloading the QGraphicsScene.
    #     """
    #     if painter.paintEngine().type() != QPaintEngine.OpenGL and \
    #         painter.paintEngine().type() != QPaintEngine.OpenGL2:
    #
    #         qWarning("OpenGLScene: drawBackground needs a QGLWidget to be set as viewport on the graphics view");
    #         return
    #     # end if
    #     painter.beginNativePainting()
    #     GL.glDisable(GL.GL_DEPTH_TEST) # disable for 2D drawing
    #     GL.glClearColor(1.0, 1.0, 1.0, 1.0)
    #     GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
    #
    #     painter.endNativePainting()
    # # end def

    # def drawBackgroundNonGL(self, painter, rect):
    #     """
    #     This method is for overloading the QGraphicsScene.
    #     """
    #     print self
    #     return QGraphicsScene.drawBackground(self, painter, rect)
    # # end def

    ### PRIVATE HELPER METHODS ###
    def _readSettings(self):
        self.settings.beginGroup("MainWindow")
        self.resize(self.settings.value("size", QSize(1100, 800)))
        self.move(self.settings.value("pos", QPoint(200, 200)))
        self.settings.endGroup()



#!/usr/bin/env python
# encoding: utf-8

from collections import defaultdict
from math import ceil
from .activesliceitem import ActiveSliceItem
from cadnano2.controllers.itemcontrollers.partitemcontroller import PartItemController
from .prexoveritem import PreXoverItem
from .strand.xoveritem import XoverNode3
from cadnano2.ui.mainwindow.svgbutton import SVGButton
from cadnano2.views import styles
from .virtualhelixitem import VirtualHelixItem
import cadnano2.util as util
from cadnano2.cadnano import app

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['QDir', 'QPointF', 'QRectF', 'Qt'])
util.qtWrapImport('QtGui', globals(), ['QBrush',
                                       'QPen'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsPathItem',
                                           'QGraphicsItem',
                                           'QGraphicsRectItem',
                                           'QInputDialog'])


_baseWidth = _bw = styles.PATH_BASE_WIDTH
_defaultRect = QRectF(0, 0, _baseWidth, _baseWidth)
_modPen = QPen(styles.bluestroke)


class ProxyParentItem(QGraphicsRectItem):
    """an invisible container that allows one to play with Z-ordering"""
    findChild = util.findChild  # for debug


class PartItem(QGraphicsRectItem):
    findChild = util.findChild  # for debug

    def __init__(self, modelPart, viewroot, activeTool, parent):
        """parent should always be pathrootitem"""
        super(PartItem, self).__init__(parent)
        self._modelPart = mP = modelPart
        self._viewroot = viewroot
        self._activeTool = activeTool
        self._activeSliceItem = ActiveSliceItem(self, mP.activeBaseIndex())
        self._activeVirtualHelixItem = None
        self._controller = PartItemController(self, mP)
        self._preXoverItems = []  # crossover-related
        self._virtualHelixHash = {}
        self._virtualHelixItemList = []
        self._vHRect = QRectF()
        self.setAcceptHoverEvents(True)
        self._initModifierRect()
        self._initResizeButtons()
        self._proxyParent = ProxyParentItem(self)
        self._proxyParent.setFlag(QGraphicsItem.GraphicsItemFlag.ItemHasNoContents)
    # end def

    def proxy(self):
        return self._proxyParent
    # end def

    def _initModifierRect(self):
        """docstring for _initModifierRect"""
        self._canShowModRect = False
        self._modRect = mR = QGraphicsRectItem(_defaultRect, self)
        mR.setPen(_modPen)
        mR.hide()
    # end def

    def _initResizeButtons(self):
        """Instantiate the buttons used to change the canvas size."""
        self._addBasesButton = SVGButton("icons:add-bases.svg", self)
        self._addBasesButton.clicked.connect(self._addBasesClicked)
        self._addBasesButton.hide()
        self._removeBasesButton = SVGButton("icons:remove-bases.svg", self)
        self._removeBasesButton.clicked.connect(self._removeBasesClicked)
        self._removeBasesButton.hide()
    # end def

    def vhItemForVH(self, vhref):
        """Returns the pathview VirtualHelixItem corresponding to vhref"""
        vh = self._modelPart.virtualHelix(vhref)
        return self._virtualHelixHash.get(vh.coord())

    ### SIGNALS ###

    ### SLOTS ###
    def partParentChangedSlot(self, sender):
        """docstring for partParentChangedSlot"""
        # print "PartItem.partParentChangedSlot"
        pass
    # end def

    def partHideSlot(self, sender):
        self.hide()
    # end def

    def partActiveVirtualHelixChangedSlot(self, part, virtualHelix):
        self.updatePreXoverItems()
    #end def

    def partDimensionsChangedSlot(self, part):
        if len(self._virtualHelixItemList) > 0:
            vhi = self._virtualHelixItemList[0]
            vhiRect = vhi.boundingRect()
            vhiHRect = vhi.handle().boundingRect()
            self._vHRect.setLeft(vhiHRect.left())
            self._vHRect.setRight(vhiRect.right())
        self.scene().views()[0].zoomToFit()
        self._activeSliceItem.resetBounds()
        self._updateBoundingRect()
    # end def

    def partRemovedSlot(self, sender):
        """docstring for partRemovedSlot"""
        self._activeSliceItem.removed()
        self.parentItem().removePartItem(self)
        scene = self.scene()
        scene.removeItem(self)
        self._modelPart = None
        self._virtualHelixHash = None
        self._virtualHelixItemList = None
        self._controller.disconnectSignals()
        self._controller = None
    # end def

    def partPreDecoratorSelectedSlot(self, sender, row, col, baseIdx):
        """docstring for partPreDecoratorSelectedSlot"""
        part = self._modelPart
        vh = part.virtualHelixAtCoord((row,col))
        vhi = self.itemForVirtualHelix(vh)
        yOffset = _bw if vh.isEvenParity() else 0
        p = QPointF(baseIdx*_bw, vhi.y() + yOffset)
        view = self.window().pathGraphicsView
        view.sceneRootItem.resetTransform()
        view.centerOn(p)
        view.zoomIn()
        self._modRect.setPos(p)
        if self._canShowModRect:
            self._modRect.show()
    # end def

    def partVirtualHelixAddedSlot(self, sender, modelVirtualHelix):
        """
        When a virtual helix is added to the model, this slot handles
        the instantiation of a virtualhelix item.
        """
        # print "PartItem.partVirtualHelixAddedSlot"
        vh = modelVirtualHelix
        vhi = VirtualHelixItem(self, modelVirtualHelix, self._viewroot, self._activeTool)
        self._virtualHelixHash[vh.coord()] = vhi
        self._virtualHelixItemList.append(vhi)
        self._setVirtualHelixItemList(self._virtualHelixItemList)
        self._updateBoundingRect()
    # end def

    def partVirtualHelixRenumberedSlot(self, sender, coord):
        """Notifies the virtualhelix at coord to change its number"""
        vh = self._virtualHelixHash[coord]
        # check for new number
        # notify VirtualHelixHandleItem to update its label
        # notify VirtualHelix to update its xovers
        # if the VirtualHelix is active, refresh prexovers
        pass
    # end def

    def partVirtualHelixResizedSlot(self, sender, coord):
        """Notifies the virtualhelix at coord to resize."""
        vh = self._virtualHelixHash[coord]
        vh.resize()
    # end def

    def partVirtualHelicesReorderedSlot(self, sender, orderedCoordList):
        """docstring for partVirtualHelicesReorderedSlot"""
        newList = self._virtualHelixItemList
        decorated = [(orderedCoordList.index(vhi.coord()), vhi)\
                        for vhi in self._virtualHelixItemList]
        decorated.sort()
        newList = [vhi for idx, vhi in decorated]
        self._setVirtualHelixItemList(newList)
    # end def

    def updatePreXoverItemsSlot(self, sender, virtualHelix):
        part = self.part()
        if virtualHelix == None:
            self.setPreXoverItemsVisible(None)
        elif part.areSameOrNeighbors(part.activeVirtualHelix(), virtualHelix):
            vhi = self.itemForVirtualHelix(virtualHelix)
            self.setActiveVirtualHelixItem(vhi)
            self.setPreXoverItemsVisible(self.activeVirtualHelixItem())
    # end def

    ### ACCESSORS ###
    def activeTool(self):
        return self._activeTool
    # end def

    def activeVirtualHelixItem(self):
        return self._activeVirtualHelixItem
    # end def

    def part(self):
        """Return a reference to the model's part object"""
        return self._modelPart
    # end def

    def document(self):
        """Return a reference to the model's document object"""
        return self._modelPart.document()
    # end def

    def removeVirtualHelixItem(self, virtualHelixItem):
        vh = virtualHelixItem.virtualHelix()
        self._virtualHelixItemList.remove(virtualHelixItem)
        del self._virtualHelixHash[vh.coord()]
        self._setVirtualHelixItemList(self._virtualHelixItemList)
        self._updateBoundingRect()

    # end def

    def itemForVirtualHelix(self, virtualHelix):
        return self._virtualHelixHash[virtualHelix.coord()]
    # end def

    def virtualHelixBoundingRect(self):
        return self._vHRect
    # end def

    def window(self):
        return self.parentItem().window()
    # end def

    ### PRIVATE METHODS ###
    def _addBasesClicked(self):
        part = self._modelPart
        step = part.stepSize()
        self._addBasesDialog = dlg = QInputDialog(self.window())
        dlg.setInputMode(QInputDialog.InputMode.IntInput)
        dlg.setIntMinimum(0)
        dlg.setIntValue(step)
        dlg.setIntMaximum(100000)
        dlg.setIntStep(step)
        dlg.setLabelText(( "Number of bases to add to the existing"\
                         + " %i bases\n(must be a multiple of %i)")\
                         % (part.maxBaseIdx(), step))
        dlg.intValueSelected.connect(self._addBasesCallback)
        dlg.open()
    # end def

    def _addBasesCallback(self, n):
        """
        Given a user-chosen number of bases to add, snap it to an index
        where index modulo stepsize is 0 and calls resizeVirtualHelices to
        adjust to that size.
        """
        part = self._modelPart
        self._addBasesDialog.intValueSelected.disconnect(self._addBasesCallback)
        del self._addBasesDialog
        maxDelta = int(n/part.stepSize()) * part.stepSize()
        part.resizeVirtualHelices(0, maxDelta)
        if app().isInMaya():
            import maya.cmds as cmds
            cmds.select(clear=True)
    # end def

    def _removeBasesClicked(self):
        """
        Determines the minimum maxBase index where index modulo stepsize == 0
        and is to the right of the rightmost nonempty base, and then resize
        each calls the resizeVirtualHelices to adjust to that size.
        """
        part = self._modelPart
        stepSize = part.stepSize()
        # first find out the right edge of the part
        idx = part.indexOfRightmostNonemptyBase()
        # next snap to a multiple of stepsize
        idx = int(ceil(float(idx+1)/stepSize))*stepSize
        # finally, make sure we're a minimum of stepSize bases
        idx = util.clamp(idx, part.stepSize(), 10000)
        delta = idx - (part.maxBaseIdx() + 1)
        if delta < 0:
            part.resizeVirtualHelices(0, delta)
            if app().isInMaya():
                import maya.cmds as cmds
                cmds.select(clear=True)
    # end def

    def _setVirtualHelixItemList(self, newList, zoomToFit=True):
        """
        Give me a list of VirtualHelixItems and I'll parent them to myself if
        necessary, position them in a column, adopt their handles, and
        position them as well.
        """
        y = 0  # How far down from the top the next PH should be
        leftmostExtent = 0
        rightmostExtent = 0

        scene = self.scene()
        vhiRect = None
        vhiHRect = None

        for vhi in newList:
            vhi.setPos(0, y)
            if not vhiRect:
                vhiRect = vhi.boundingRect()
                step = vhiRect.height() + styles.PATH_HELIX_PADDING
            # end if

            # get the VirtualHelixHandleItem
            vhiH = vhi.handle()
            if vhiH.parentItem() != self._viewroot._vhiHSelectionGroup:
                vhiH.setParentItem(self)

            if not vhiHRect:
                vhiHRect = vhiH.boundingRect()

            vhiH.setPos(-2 * vhiHRect.width(), y + (vhiRect.height() - vhiHRect.height()) / 2)

            leftmostExtent = min(leftmostExtent, -2 * vhiHRect.width())
            rightmostExtent = max(rightmostExtent, vhiRect.width())
            y += step
            self.updateXoverItems(vhi)
        # end for
        self._vHRect = QRectF(leftmostExtent, -40, -leftmostExtent + rightmostExtent, y + 40)
        self._virtualHelixItemList = newList
        if zoomToFit:
            self.scene().views()[0].zoomToFit()
    # end def

    def _updateBoundingRect(self):
        """
        Updates the bounding rect to the size of the childrenBoundingRect,
        and refreshes the addBases and removeBases buttons accordingly.

        Called by partVirtualHelixAddedSlot, partDimensionsChangedSlot, or
        removeVirtualHelixItem.
        """
        self.setPen(QPen(Qt.PenStyle.NoPen))
        self.setRect(self.childrenBoundingRect())
        # move and show or hide the buttons if necessary
        addButton = self._addBasesButton
        rmButton = self._removeBasesButton
        if len(self._virtualHelixItemList) > 0:
            addRect = addButton.boundingRect()
            rmRect = rmButton.boundingRect()
            x = self._vHRect.right()
            y = -styles.PATH_HELIX_PADDING
            addButton.setPos(x, y)
            rmButton.setPos(x-rmRect.width(), y)
            addButton.show()
            rmButton.show()
        else:
            addButton.hide()
            rmButton.hide()
    # end def

    ### PUBLIC METHODS ###
    def setModifyState(self, bool):
        """Hides the modRect when modify state disabled."""
        self._canShowModRect = bool
        if bool == False:
            self._modRect.hide()

    def getOrderedVirtualHelixList(self):
        """Used for encoding."""
        ret = []
        for vhi in self._virtualHelixItemList:
            ret.append(vhi.coord())
        return ret
    # end def

    def numberOfVirtualHelices(self):
        return len(self._virtualHelixItemList)
    # end def

    def reorderHelices(self, first, last, indexDelta):
        """
        Reorder helices by moving helices _pathHelixList[first:last]
        by a distance delta in the list. Notify each PathHelix and
        PathHelixHandle of its new location.
        """
        vhiList = self._virtualHelixItemList
        helixNumbers = [vhi.number() for vhi in vhiList]
        firstIndex = helixNumbers.index(first)
        lastIndex = helixNumbers.index(last) + 1

        if indexDelta < 0:  # move group earlier in the list
            newIndex = max(0, indexDelta + firstIndex)
            newList = vhiList[0:newIndex] +\
                                vhiList[firstIndex:lastIndex] +\
                                vhiList[newIndex:firstIndex] +\
                                vhiList[lastIndex:]
        # end if
        else:  # move group later in list
            newIndex = min(len(vhiList), indexDelta + lastIndex)
            newList = vhiList[:firstIndex] +\
                                 vhiList[lastIndex:newIndex] +\
                                 vhiList[firstIndex:lastIndex] +\
                                 vhiList[newIndex:]
        # end else

        # call the method to move the items and store the list
        self._setVirtualHelixItemList(newList, zoomToFit=False)
    # end def

    def setActiveVirtualHelixItem(self, newActiveVHI):
        if newActiveVHI != self._activeVirtualHelixItem:
            self._activeVirtualHelixItem = newActiveVHI
            # self._modelPart.setActiveVirtualHelix(newActiveVHI.virtualHelix())
    # end def

    def setPreXoverItemsVisible(self, virtualHelixItem):
        """
        self._preXoverItems list references prexovers parented to other
        PathHelices such that only the activeHelix maintains the list of
        visible prexovers
        """
        vhi = virtualHelixItem

        if vhi == None:
            if self._preXoverItems:
                # clear all PreXoverItems
                list(map(PreXoverItem.remove, self._preXoverItems))
                self._preXoverItems = []
            return

        vh = vhi.virtualHelix()
        partItem = self
        part = self.part()
        idx = part.activeVirtualHelixIdx()

        # clear all PreXoverItems
        list(map(PreXoverItem.remove, self._preXoverItems))
        self._preXoverItems = []

        potentialXovers = part.potentialCrossoverList(vh, idx)
        for neighbor, index, strandType, isLowIdx in potentialXovers:
            # create one half
            neighborVHI = self.itemForVirtualHelix(neighbor)
            pxi = PreXoverItem(vhi, neighborVHI, index, strandType, isLowIdx)
            # add to list
            self._preXoverItems.append(pxi)
            # create the complement
            pxi = PreXoverItem(neighborVHI, vhi, index, strandType, isLowIdx)
            # add to list
            self._preXoverItems.append(pxi)
        # end for
    # end def

    def updatePreXoverItems(self):
        self.setPreXoverItemsVisible(self.activeVirtualHelixItem())
    # end def

    def updateXoverItems(self, virtualHelixItem):
        for item in virtualHelixItem.childItems():
            if isinstance(item, XoverNode3):
                item.refreshXover()
     # end def

    def updateStatusBar(self, statusString):
        """Shows statusString in the MainWindow's status bar."""
        self.window().statusBar().showMessage(statusString)

    ### COORDINATE METHODS ###
    def keyPanDeltaX(self):
        """How far a single press of the left or right arrow key should move
        the scene (in scene space)"""
        vhs = self._virtualHelixItemList
        return vhs[0].keyPanDeltaX() if vhs else 5
    # end def

    def keyPanDeltaY(self):
        """How far an an arrow key should move the scene (in scene space)
        for a single press"""
        vhs = self._virtualHelixItemList
        if not len(vhs) > 1:
            return 5
        dy = vhs[0].pos().y() - vhs[1].pos().y()
        dummyRect = QRectF(0, 0, 1, dy)
        return self.mapToScene(dummyRect).boundingRect().height()
    # end def

    ### TOOL METHODS ###
    def hoverMoveEvent(self, event):
        """
        Parses a mouseMoveEvent to extract strandSet and base index,
        forwarding them to approproate tool method as necessary.
        """
        activeTool = self._activeTool()
        toolMethodName = str(activeTool) + "HoverMove"
        if hasattr(self, toolMethodName):
            getattr(self, toolMethodName)(event.pos())
    # end def

    def pencilToolHoverMove(self, pt):
        """Pencil the strand is possible."""
        partItem = self
        activeTool = self._activeTool()
        if not activeTool.isFloatingXoverBegin():
            tempXover = activeTool.floatingXover()
            tempXover.updateFloatingFromPartItem(self, pt)
    # end def



from os.path import basename
from cadnano2.model.enum import StrandType


def legacy_dict_from_doc(document, fname, helixOrderList):
    part = document.selectedPart()
    numBases = part.maxBaseIdx()+1

    # iterate through virtualhelix list
    vhList = []
    for row, col in helixOrderList:
        vh = part.virtualHelixAtCoord((row, col))
        # insertions and skips
        insertionDict = part.insertions()[(row, col)]
        insts = [0 for i in range(numBases)]
        skips = [0 for i in range(numBases)]
        for idx, insertion in insertionDict.items():
            if insertion.isSkip():
                skips[idx] = insertion.length()
            else:
                insts[idx] = insertion.length()
        # colors
        stapColors = []
        stapStrandSet = vh.stapleStrandSet()
        for strand in stapStrandSet:
            if strand.connection5p() == None:
                c = str(strand.oligo().color())[1:]  # drop the hash
                stapColors.append([strand.idx5Prime(), int(c, 16)])

        vhDict = {"row": row,
                  "col": col,
                  "num": vh.number(),
                  "scaf": vh.getLegacyStrandSetArray(StrandType.Scaffold),
                  "stap": vh.getLegacyStrandSetArray(StrandType.Staple),
                  "loop": insts,
                  "skip": skips,
                  "scafLoop": [],
                  "stapLoop": [],
                  "stap_colors": stapColors}
        vhList.append(vhDict)
    bname = basename(str(fname))
    obj = {"name": bname, "vstrands": vhList}
    return obj



from urllib.request import Request, urlopen
from urllib.error import URLError, HTTPError
import sys
from os.path import basename, dirname, splitext, exists
import os
import shutil
import tarfile

def fetchFile(filename, baseurl, filemode='b', filetype='gz', filepath=None):
    """
    
    filepath - is optional file path location to store the fetched file
    """
    # create the url and the request
    url = baseurl + '/' + filename
    request = Request(url)

    # Open the url
    try:
        f_url = urlopen(request)
        print("downloading " + url)

        # Open our local file for writing
        f_dest = open(filename, "w" + filemode)
        
        # Write to our local file
        f_dest.write(f_url.read())
        f_dest.close()
        
    # handle errors
    except HTTPError as e:
        print("HTTP Error:", e.code , url)
    except URLError as e:
        print("URL Error:", e.reason , url)
    
    filename_out = filename
    # unzip if possible
    if filetype == 'gz':
        
        # get the extracted folder name
        filename_out = splitext(filename)[0]
        temp = splitext(filename_out)
        if temp[1] == '.tar':
            filename_out = temp[0]
        
        # open the archive
        try:
            f_zip= tarfile.open(filename, mode='r')
        except tarfile.ReadError as e:
            print("unable to read archive", e.code)
        print("extracting " + filename_out)
        try:
            if filepath:
                # remove existing folder
                if os.path.exists(filepath + '/' + filename_out):
                    print("file exists")
                    shutil.rmtree(filepath + '/' + filename_out)
                else:
                    print("file does not exist", filename_out)
                f_zip.extractall(path=filepath)
            else:
                # remove existing folder
                if os.path.exists(filename_out):
                    print("file exists")
                    shutil.rmtree(filename_out)
                else:
                    print("file does not exist", filename_out)
                f_zip.extractall()
        except tarfile.ExtractError as e:
            print("unable to extract archive", e.code)
        f_zip.close()
        # remove the archive
        print("removing the downloaded archive", filename)
        os.remove(filename)
        print("done")
    return filename_out
# end def

if __name__ == '__main__':
    argv = sys.argv
    url = argv[1]
    filename = basename(url)
    base_url = dirname(url)
    fetchFile(filename, base_url)




"""
cadnanoqt
Created by Jonathan deWerd on 2012-01-11.
"""
import sys, os
from . import cadnano, util
from code import interact
util.qtWrapImport('QtWidgets', globals(), ['QApplication'])
util.qtWrapImport('QtGui', globals(),  ['QIcon', 'QUndoGroup'])
util.qtWrapImport('QtCore', globals(), ['QObject', 'QCoreApplication', 'Qt',
                                        'QEventLoop', 'pyqtSignal'])


class CadnanoQt(QObject):
    dontAskAndJustDiscardUnsavedChanges = False
    shouldPerformBoilerplateStartupScript = False
    documentWasCreatedSignal = pyqtSignal(object)  # doc
    documentWindowWasCreatedSignal = pyqtSignal(object, object)  # doc, window

    def __init__(self, argv):
        self.argv = argv
        if QCoreApplication.instance() == None:
            self.qApp = QApplication(argv)
            assert(QCoreApplication.instance() != None)
            self.qApp.setOrganizationDomain("cadnano.org")
        else:
            raise Exception
            # self.qApp = qApp
        super(CadnanoQt, self).__init__()
        from .views.preferences import Preferences
        self.prefs = Preferences()
        self.qApp.setWindowIcon(QIcon('ui/mainwindow/images/cadnano2-app-icon.png'))
        self.documentControllers = set()  # Open documents
        self.activeDocument = None
        self.vh = {}  # Newly created VirtualHelix register here by idnum.
        self.vhi = {}
        self.partItem = None
        self.sharedApp = self

    def ignoreEnv(self):
        return os.environ.get('CADNANO_IGNORE_ENV_VARS_EXCEPT_FOR_ME', False)

    def finishInit(self):
        self.d = self.newDocument(isFirstNewDoc=True)
        # os.environ['CADNANO_DISCARD_UNSAVED'] = 'True' ## added by Nick 
        if os.environ.get('CADNANO_DISCARD_UNSAVED', False) and not self.ignoreEnv():
            self.sharedApp.dontAskAndJustDiscardUnsavedChanges = True
        if os.environ.get('CADNANO_DEFAULT_DOCUMENT', False) and not self.ignoreEnv():
            self.sharedApp.shouldPerformBoilerplateStartupScript = True
        # cadnano.loadAllPlugins()
        if "-i" in self.argv:
            print("Welcome to cadnano's debug mode!")
            print("Some handy locals:")
            print("\ta\tcadnano.app() (the shared cadnano application object)")
            print("\td()\tthe last created Document")
            def d():
                return self.d

            print("\tw()\tshortcut for d().controller().window()")
            def w():
                return self.d.controller().window()

            print("\tp()\tshortcut for d().selectedPart()")
            def p():
                return self.d.selectedPart()

            print("\tpi()\tthe PartItem displaying p()")
            def pi():
                return w().pathroot.partItemForPart(p())

            print("\tvh(i)\tshortcut for p().virtualHelix(i)")
            def vh(vhref):
                return p().virtualHelix(vhref)

            print("\tvhi(i)\tvirtualHelixItem displaying vh(i)")
            def vhi(vhref):
                partitem = pi()
                vHelix = vh(vhref)
                return partitem.vhItemForVH(vHelix)

            print("\tquit()\tquit (for when the menu fails)")
            print("\tgraphicsItm.findChild()  see help(pi().findChild)")
            interact('', local={'a':self, 'd':d, 'w':w,\
                                'p':p, 'pi':pi, 'vh':vh, 'vhi':vhi,\
                                })
        # else:
        #     self.exec_()

    def isInMaya(self):
        return "Maya" in QCoreApplication.instance().applicationName()

    def isGui(self):
        return True

    def destroyApp(self):
        """ Destroy the QApplication.

        Do not set `self.qApp = None` in this method.
        Do it external to the CadnanoQt class
        """
        global decodeFile
        global Document
        global DocumentController
        # print("documentWasCreatedSignal", self.documentWasCreatedSignal)
        # if self.documentControllers:
        #     self.documentWasCreatedSignal.disconnect(self.wirePrefsSlot)
        decodeFile = None
        Document = None
        DocumentController = None
        self.documentControllers.clear()
        self.qApp.quit()
        sys.exit(0)
    # end def

    def exec_(self):
        if hasattr(self, 'qApp'):
            self.mainEventLoop = QEventLoop()
            self.mainEventLoop.exec()
            #self.qApp.exec_()

    def newDocument(self, isFirstNewDoc=False):
        from cadnano2.controllers.documentcontroller import DocumentController
        defaultFile = os.environ.get('CADNANO_DEFAULT_DOCUMENT', None)
        if defaultFile and isFirstNewDoc:
            defaultFile = path.expanduser(defaultFile)
            defaultFile = path.expandvars(defaultFile)
            dc = DocumentController()
            doc = dc.document()
            from cadnano2.model.io.decoder import decode
            decode(doc, file(defaultFile).read())
            print("Loaded default document: %s" % doc)
        else:
            docCtrlrCount = len(self.documentControllers)
            if docCtrlrCount == 0:  # first dc
                # dc adds itself to app.documentControllers
                dc = DocumentController()
            elif docCtrlrCount == 1:  # dc already exists
                dc = list(self.documentControllers)[0]
                dc.newDocument()  # tell it to make a new doucment
        return dc.document()

    def prefsClicked(self):
        self.prefs.showDialog()



import cadnano2.util as util
# from cadnano2.controllers.itemcontrollers.abstractitemcontroller import AbstractItemController


class AbstractStrandItemController(object):
    def __init__(self, strandItem, modelStrand):
        """
        Do not call connectSignals here.  subclasses
        will install two sets of signals.
        """
        if self.__class__ == AbstractStrandItemController:
            e = "AbstractStrandItemController should be subclassed."
            raise NotImplementedError(e)
        self._strandItem = strandItem
        self._modelStrand = modelStrand
        self._modelOligo = modelStrand.oligo()
    # end def

    def reconnectOligoSignals(self):
        self.disconnectOligoSignals()
        self.connectOligoSignals()
    # end def

    def connectSignals(self):
        """Connects modelStrant signals to strandItem slots."""
        mS = self._modelStrand
        sI = self._strandItem

        AbstractStrandItemController.connectOligoSignals(self)
        mS.strandHasNewOligoSignal.connect(sI.strandHasNewOligoSlot)
        mS.strandRemovedSignal.connect(sI.strandRemovedSlot)

        mS.strandInsertionAddedSignal.connect(sI.strandInsertionAddedSlot)
        mS.strandInsertionChangedSignal.connect(sI.strandInsertionChangedSlot)
        mS.strandInsertionRemovedSignal.connect(sI.strandInsertionRemovedSlot)
        mS.strandDecoratorAddedSignal.connect(sI.strandDecoratorAddedSlot)
        mS.strandDecoratorChangedSignal.connect(sI.strandDecoratorChangedSlot)
        mS.strandDecoratorRemovedSignal.connect(sI.strandDecoratorRemovedSlot)
        mS.strandModifierAddedSignal.connect(sI.strandModifierAddedSlot)
        mS.strandModifierChangedSignal.connect(sI.strandModifierChangedSlot)
        mS.strandModifierRemovedSignal.connect(sI.strandModifierRemovedSlot)

        mS.selectedChangedSignal.connect(sI.selectedChangedSlot)
    # end def

    def connectOligoSignals(self):
        sI = self._strandItem
        mO = self._modelStrand.oligo()
        self._modelOligo = mO
        mO.oligoAppearanceChangedSignal.connect(sI.oligoAppearanceChangedSlot)
    # end def

    def disconnectSignals(self):
        mS = self._modelStrand
        sI = self._strandItem

        AbstractStrandItemController.disconnectOligoSignals(self)
        mS.strandHasNewOligoSignal.disconnect(sI.strandHasNewOligoSlot)
        mS.strandRemovedSignal.disconnect(sI.strandRemovedSlot)

        mS.strandInsertionAddedSignal.disconnect(sI.strandInsertionAddedSlot)
        mS.strandInsertionChangedSignal.disconnect(sI.strandInsertionChangedSlot)
        mS.strandInsertionRemovedSignal.disconnect(sI.strandInsertionRemovedSlot)
        mS.strandDecoratorAddedSignal.disconnect(sI.strandDecoratorAddedSlot)
        mS.strandDecoratorChangedSignal.disconnect(sI.strandDecoratorChangedSlot)
        mS.strandDecoratorRemovedSignal.disconnect(sI.strandDecoratorRemovedSlot)
        mS.strandModifierAddedSignal.disconnect(sI.strandModifierAddedSlot)
        mS.strandModifierChangedSignal.disconnect(sI.strandModifierChangedSlot)
        mS.strandModifierRemovedSignal.disconnect(sI.strandModifierRemovedSlot)

        mS.selectedChangedSignal.disconnect(sI.selectedChangedSlot)
    # end def

    def disconnectOligoSignals(self):
        sI = self._strandItem
        mO = self._modelOligo
        mO.oligoAppearanceChangedSignal.disconnect(sI.oligoAppearanceChangedSlot)
    # end def



import json
from .legacydecoder import import_legacy_dict
from cadnano2.ui.dialogs.ui_latticetype import Ui_LatticeType
import cadnano2.util as util
import cadnano2.cadnano as cadnano
if cadnano.app().isGui():  # headless:
    from cadnano2.ui.dialogs.ui_latticetype import Ui_LatticeType
    util.qtWrapImport('QtWidgets', globals(),  ['QDialog', 'QDialogButtonBox'])


def decode(document, string):
    if cadnano.app().isGui():
        # from ui.dialogs.ui_latticetype import Ui_LatticeType
        # util.qtWrapImport('QtGui', globals(),  ['QDialog', 'QDialogButtonBox'])
        dialog = QDialog()
        dialogLT = Ui_LatticeType()  # reusing this dialog, should rename
        dialogLT.setupUi(dialog)

    # try:  # try to do it fast
    #     try:
    #         import cjson
    #         packageObject = cjson.decode(string)
    #     except:  # fall back to if cjson not available or on decode error
    #         packageObject = json.loads(string)
    # except ValueError:
    #     dialogLT.label.setText("Error decoding JSON object.")
    #     dialogLT.buttonBox.setStandardButtons(QDialogButtonBox.Ok)
    #     dialog.exec()
    #     return
    packageObject = json.loads(string)

    if packageObject.get('.format', None) != 'caDNAno2':
        import_legacy_dict(document, packageObject)


from cadnano2.controllers.viewrootcontroller import ViewRootController
from .partitem import PartItem
import cadnano2.util as util
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'QObject'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsRectItem'])


class SliceRootItem(QGraphicsRectItem):
    """
    PathRootItem is the root item in the PathView. It gets added directly
    to the pathscene by DocumentWindow. It receives two signals
    (partAddedSignal and selectedPartChangedSignal) via its ViewRootController.

    PathRootItem must instantiate its own controller to receive signals
    from the model.
    """
    def __init__(self, rect, parent, window, document):
        super(SliceRootItem, self).__init__(rect, parent)
        self._window = window
        self._document = document
        self._controller = ViewRootController(self, document)
        self._instanceItems = {}

    ### SIGNALS ###

    ### SLOTS ###
    def partAddedSlot(self, sender, modelPart):
        """
        Receives notification from the model that a part has been added.
        Views that subclass AbstractView should override this method.
        """
        self._modelPart = modelPart
        partItem = PartItem(modelPart, parent=self)
        self._instanceItems[partItem] = partItem
        self.setModifyState(self._window.actionModify.isChecked())
    # end def

    def selectedChangedSlot(self, itemDict):
        """docstring for selectedChangedSlot"""
        pass
    # end def
    
    def selectionFilterChangedSlot(self, filterNameList):
        pass
    # end def
    
    def clearSelectionsSlot(self, doc):
        self.scene().views()[0].clearSelectionLockAndCallbacks()
    # end def
    
    def resetRootItemSlot(self, doc):
        pass
    # end def

    ### ACCESSORS ###
    def sliceToolManager(self):
        """docstring for sliceToolManager"""
        return self._window.sliceToolManager
    # end def

    def window(self):
        return self._window
    # end def

    ### METHODS ###
    def removePartItem(self, partItem):
        del self._instanceItems[partItem]
    # end def

    def resetDocumentAndController(self, document):
        """docstring for resetDocumentAndController"""
        self._document = document
        self._controller = ViewRootController(self, document)
        if len(self._instanceItems) > 0:
            raise ImportError
    # end def

    def setModifyState(self, bool):
        """docstring for setModifyState"""
        for partItem in self._instanceItems:
            partItem.setModifyState(bool)
    # end def



# Copyright 2011 Autodesk, Inc.  All rights reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# http://www.opensource.org/licenses/mit-license.php

"""
mayaObjectManager.py
Created by Alex Tessier on 2011-08
"""
from cadnano2.cadnano import app
import maya.cmds as cmds
import cadnano2.util as util
import maya.OpenMaya as OpenMaya


class Mom(object):
    """
    A singleton manager for tracking maya to cadnano and reverse lookups of
    strand and pre-decorators
    """
    _instance = None
    strandCount = 0
    selectionCountCache = 0

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Mom, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def myId(self):
        return id(self)

    # uses DNACylinderShape as as the key, stores strand objects
    mayaToCn = {}
    # uses strand objects as the key, stores a list of maya nodes
    cnToMaya = {}
    # uses stapleModIndicatorMesh% objects as the key,
    # stores a object with (virualHelixItem object, baseNumber, strand object)
    decoratorToVirtualHelixItem = {}
    # uses strand object as the key, stores stand id
    idStrandMapping = {}
    # Selection
    ignoreExternalSelectionSignal = False

    # Selection boxes
    selectionBox = cmds.polyCube(
                        constructionHistory=False,
                        createUVs=0,
                        object=True
                        )[0]
    cmds.hide(selectionBox)
    cmds.setAttr(selectionBox + ".overrideEnabled", True)
    cmds.setAttr(selectionBox + ".overrideDisplayType", 2)  # reference
    selectionBoxShader = "SelectionBoxShader"

    selectionBoxShader = cmds.shadingNode('lambert',
                                        asShader=True,
                                        name=selectionBoxShader)
    cmds.sets(name="%sSG" % selectionBoxShader,
                renderable=True,
                noSurfaceShader=True,
                empty=True)
    cmds.connectAttr("%s.outColor" % selectionBoxShader,
                    "%sSG.surfaceShader" % selectionBoxShader)
    cmds.setAttr("%s.transparency" % selectionBoxShader,
                    0.65, 0.65, 0.65, type="double3")
    cmds.setAttr("%s.incandescence" % selectionBoxShader,
                    0.5, 0.5, 0.5, type="double3")
    cmds.sets(
            selectionBox,
            forceElement="%sSG" % selectionBoxShader)

    # Endpoint Selection boxes
    epSelectionBox = cmds.polyCube(
                        constructionHistory=False,
                        createUVs=0,
                        object=True
                        )[0]
    cmds.hide(epSelectionBox)
    cmds.setAttr(epSelectionBox + ".overrideEnabled", True)
    cmds.setAttr(epSelectionBox + ".overrideDisplayType", 2)  # reference
    epSelectionBoxShader = "epSelectionBoxShader"

    epSelectionBoxShader = cmds.shadingNode('lambert',
                                        asShader=True,
                                        name=epSelectionBoxShader)
    cmds.sets(name="%sSG" % epSelectionBoxShader,
            renderable=True,
            noSurfaceShader=True,
            empty=True)
    cmds.connectAttr("%s.outColor" % epSelectionBoxShader,
                    "%sSG.surfaceShader" % epSelectionBoxShader)
    cmds.setAttr("%s.color" % epSelectionBoxShader,
                    0.8, 0.2, 0.2, type="double3")
    cmds.setAttr("%s.transparency" % epSelectionBoxShader,
                    0.75, 0.75, 0.75, type="double3")
    cmds.setAttr("%s.incandescence" % epSelectionBoxShader,
                    0.5, 0.5, 0.5, type="double3")
    cmds.sets(
            epSelectionBox,
            forceElement="%sSG" % epSelectionBoxShader)

    # MayaNames
    helixTransformName = "DNAShapeTransform_"
    helixNodeName = "HalfCylinderHelixNode_"
    helixMeshName = "DNACylinderShape_"
    helixShaderName = "DNAStrandShader_"
    decoratorTransformName = "stapleDecoratorTransform_"
    decoratorNodeName = "spStapleModIndicator_"
    decoratorMeshName = "stapleModIndicatorMesh_"
    decoratorShaderName = "stapleModeIndicatorShader_"

    def getNodeFromName(self, NodeName):
        #print "getNodeFromName ", NodeName
        selList = OpenMaya.MSelectionList()
        dependNode = 0
        if(cmds.objExists(NodeName)):
            selList.add(NodeName)
            dependNode = OpenMaya.MObject()
            selList.getDependNode(0, dependNode)
        return dependNode

    def getSelectionBox(self):
        return self.selectionBox

    def strandsSelected(self, listNames, value=(True, True)):
        if self.ignoreExternalSelectionSignal:
            return

        self.ignoreExternalSelectionSignal = True
        strandList = []
        for nodeName in listNames:
            if(nodeName in self.mayaToCn):
                strandList.append(self.mayaToCn[nodeName])
        doc = app().activeDocument
        # # XXX [SB] THIS IS A HACK, should not need to do this!!!
        # doc.win.pathroot.clearStrandSelections()

        doc.win.solidroot.selectedChanged(strandList, value)
        self.ignoreExternalSelectionSignal = False

    def staplePreDecoratorSelected(self, listNames):
        """
        Callback function that is called from mayaSelectionContext when a
        PreDecorator geometry is called, notifies the Part Model of this
        event. XXX - [SB] In the future we should clean up this interaction.
        """
        if(len(listNames) > 1):
            # If we have more than one PreDecorator Selected, deselect all but
            # the last one
            cmds.select(listNames[0:len(listNames) - 1], deselect=True)

        selectionList = []

        for name in listNames:
            if name in self.decoratorToVirtualHelixItem:
                (virtualHelixItem, baseIdx, strand) = \
                                        self.decoratorToVirtualHelixItem[name]
                selectionList.append((virtualHelixItem.row(),
                                                virtualHelixItem.col(),
                                                baseIdx))
        if len(selectionList) == 0 and \
           self.selectionCountCache == 0:
            # a dumb cache check to prevent deselection to be broadcasted too
            # many times, but could cause problems
            return

        # XXX - [SB] we want to only send the signal to "active part" but
        # not sure how to get that
        for doc in app().documentControllers:
            if doc.win.actionModify.isChecked():
                for partItem in doc.win.solidroot.partItems():
                    partModel = partItem.part()
                    partModel.selectPreDecorator(selectionList)
        self.selectionCountCache = len(selectionList)

    def removeDecoratorMapping(self, id):
        """Remove all mappings related to Pre-Decorators"""
        key1 = "%s%s" % (self.decoratorMeshName, id)
        key2 = "%s%s" % (self.decoratorTransformName, id)
        if key1 in self.decoratorToVirtualHelixItem:
            del self.decoratorToVirtualHelixItem[key1]
        if key2 in self.decoratorToVirtualHelixItem:
            del self.decoratorToVirtualHelixItem[key2]

    def removeIDMapping(self, id, strand):
        """Remove all mappings related to Strands"""
        key1 = "%s%s" % (self.helixMeshName, id)
        key2 = "%s%s" % (self.helixNodeName, id)
        if key1 in self.mayaToCn:
            del self.mayaToCn[key1]
        if key2 in self.mayaToCn:
            del self.mayaToCn[key2]
        self.deleteStrandMayaID(strand)

    def strandMayaID(self, strand):
        """
        Get a Maya ID for a given strand, if one does not exits, it
        will create a new one.
        """
        if(strand in self.idStrandMapping):
            return self.idStrandMapping[strand]
        else:
            self.strandCount += 1
            # XXX [SB+AT] NOT THREAD SAFE
            while cmds.objExists("%s%s" %
                                 (self.helixTransformName,
                                  self.strandCount)):
                self.strandCount += 1
            val = "%d" % self.strandCount
            self.idStrandMapping[strand] = val
            # print self.idStrandMapping
            return val
    # end def

    def deleteStrandMayaID(self, strand):
        """
        Removes just the Cadnano -> Maya mapping.
        It is called from removeIDMapping function
        """
        if strand in self.cnToMaya:
            del self.cnToMaya[strand]
        if strand in self.idStrandMapping:
            del self.idStrandMapping[strand]

    def updateSelectionBoxes(self):
        selectedItems = cmds.ls(self.helixTransformName + "*", selection=True)
        if selectedItems:
            bbox = cmds.exactWorldBoundingBox(selectedItems)
            cmds.setAttr(
                    self.selectionBox + ".scale",
                    bbox[3] - bbox[0],
                    bbox[4] - bbox[1],
                    bbox[5] - bbox[2],
                    type="double3")
            cmds.setAttr(
                    self.selectionBox + ".translate",
                    (bbox[0] + bbox[3]) / 2,
                    (bbox[1] + bbox[4]) / 2,
                    (bbox[2] + bbox[5]) / 2,
                    type="double3")
            cmds.showHidden(self.selectionBox)

            selectionDict = app().activeDocument.document().selectionDict()

            frontEp = 0
            backEp = 0

            for strandSetDict in selectionDict.values():
                for strand, value in strandSetDict.items():
                    frontEp += int(value[0])
                    backEp += int(value[1])

            if (frontEp == 0) ^ (backEp == 0):
                epBoundBox = None
                for strandSetDict in selectionDict.values():
                    for strand, value in strandSetDict.items():
                        # XXX The following line is a work around for broken
                        # path selection in model, remove when fixed
                        if not strand in self.cnToMaya:
                            continue
                        helixNode = self.cnToMaya[strand]
                        boundBox = cmds.exactWorldBoundingBox(helixNode[1])
                        # might be better to get the rise this way...
                        #n = OpenMaya.MFnDependencyNode(helixNode)
                        #risePlug = n.findPlug("rise")
                        #rise = risePlug.asDouble()
                        # but this works too
                        idxL, idxH = strand.idxs()
                        rise = (boundBox[5] - boundBox[2]) / (idxH - idxL + 1)
                        if frontEp == 0:
                            boundBox[5] = boundBox[2] + rise
                        elif backEp == 0:
                            boundBox[2] = boundBox[5] - rise
                        if epBoundBox == None:
                            epBoundBox = boundBox
                        else:
                            # union the boxes
                            epBoundBox[0] = min(epBoundBox[0], boundBox[0])
                            epBoundBox[1] = min(epBoundBox[1], boundBox[1])
                            epBoundBox[2] = min(epBoundBox[2], boundBox[2])
                            epBoundBox[3] = max(epBoundBox[3], boundBox[3])
                            epBoundBox[4] = max(epBoundBox[4], boundBox[4])
                            epBoundBox[5] = max(epBoundBox[5], boundBox[5])

                # XXX The following line is a work around for broken
                # path selection in model, remove when fixed
                if not epBoundBox == None:

                    cmds.showHidden(self.epSelectionBox)
                    cmds.setAttr(
                            self.epSelectionBox + ".scale",
                            epBoundBox[3] - epBoundBox[0],
                            epBoundBox[4] - epBoundBox[1],
                            epBoundBox[5] - epBoundBox[2],
                            type="double3")
                    cmds.setAttr(
                            self.epSelectionBox + ".translate",
                            (epBoundBox[0] + epBoundBox[3]) / 2,
                            (epBoundBox[1] + epBoundBox[4]) / 2,
                            (epBoundBox[2] + epBoundBox[5]) / 2,
                            type="double3")

            else:
                cmds.hide(self.epSelectionBox)

        else:
            cmds.hide(self.selectionBox)
            cmds.hide(self.epSelectionBox)



"""
xoveritem.py
Created by Nick on 2011-05-25.
"""

from cadnano2.views import styles
import cadnano2.util as util

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['QPointF', 'QRectF', 'Qt', 'QEvent'])
util.qtWrapImport('QtGui', globals(), ['QBrush',
                                       'QFont',
                                       'QPen',
                                       'QPolygonF',
                                       'QPainterPath',
                                       'QColor',
                                       'QFontMetrics'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsItem',
                                           'QGraphicsSimpleTextItem',
                                           'QGraphicsRectItem',
                                           'QGraphicsPathItem'])

_baseWidth = styles.PATH_BASE_WIDTH
_toHelixNumFont = styles.XOVER_LABEL_FONT
# precalculate the height of a number font.  Assumes a fixed font
# and that only numbers will be used for labels
_fm = QFontMetrics(_toHelixNumFont)
_enabbrush = QBrush(Qt.BrushStyle.SolidPattern)  # Also for the helix number label
_nobrush = QBrush(Qt.BrushStyle.NoBrush)
# _rect = QRectF(0, 0, baseWidth, baseWidth)
_xScale = styles.PATH_XOVER_LINE_SCALE_X  # control point x constant
_yScale = styles.PATH_XOVER_LINE_SCALE_Y  # control point y constant
_rect = QRectF(0, 0, _baseWidth, _baseWidth)


class XoverNode3(QGraphicsRectItem):
    """
    This is a QGraphicsRectItem to allow actions and also a
    QGraphicsSimpleTextItem to allow a label to be drawn
    """
    def __init__(self, virtualHelixItem, xoverItem, strand3p, idx):
        super(XoverNode3, self).__init__(virtualHelixItem)
        self._vhi = virtualHelixItem
        self._xoverItem = xoverItem
        self._idx = idx
        self._isOnTop = virtualHelixItem.isStrandOnTop(strand3p)
        self._isDrawn5to3 = strand3p.strandSet().isDrawn5to3()
        self._strandType = strand3p.strandSet().strandType()

        self.setPartnerVirtualHelix(strand3p)

        self.setPen(QPen(Qt.PenStyle.NoPen))
        self._label = None
        self.setPen(QPen(Qt.PenStyle.NoPen))
        self.setBrush(_nobrush)
        self.setRect(_rect)
        self.setZValue(styles.ZXOVERITEM)
    # end def

    ### EVENT HANDLERS ###
    def mousePressEvent(self, event):
        """
        Parses a mousePressEvent to extract strandSet and base index,
        forwarding them to approproate tool method as necessary.
        """
        self.scene().views()[0].addToPressList(self)
        self._vhi.setActive(self._idx)
        xoi = self._xoverItem
        toolMethodName = str(xoi.activeTool()) + "MousePress"
        if hasattr(xoi, toolMethodName):
            getattr(xoi, toolMethodName)()
    # end def

    def customMouseRelease(self, event):
        pass
    # end def

    def virtualHelix(self):
        return self._vhi.virtualHelix()
    # end def

    def strandType(self):
        return self._strandType
    # end def

    def refreshXover(self):
        self._xoverItem.refreshXover()
    # end def

    def setPartnerVirtualHelix(self,strand):
        if strand.connection5p():
            self._partnerVirtualHelix = strand.connection5p().virtualHelix()
        else:
            self._partnerVirtualHelix = None
    # end def

    def idx(self):
        return self._idx
    # end def

    def setIdx(self, idx):
        self._idx = idx
    # end def

    def virtualHelixItem(self):
        return self._vhi
    # end def

    def point(self):
        return self._vhi.upperLeftCornerOfBaseType(self._idx, self._strandType)
    # end def

    def floatPoint(self):
        pt = self.pos()
        return pt.x(), pt.y()
    # end def

    def isOnTop(self):
        return self._isOnTop
    # end def

    def isDrawn5to3(self):
        return self._isDrawn5to3
    # end def

    def updatePositionAndAppearance(self):
        """
        Sets position by asking the VirtualHelixItem
        Sets appearance by choosing among pre-defined painterpaths (from
        normalstrandgraphicsitem) depending on drawing direction.
        """
        self.setPos(*self.point())
        # We can only expose a 5' end. But on which side?
        isLeft = True if self._isDrawn5to3 else False
        self._updateLabel(isLeft)
    # end def

    def updateConnectivity(self):
        isLeft = True if self._isDrawn5to3 else False
        self._updateLabel(isLeft)
    # end def

    def remove(self):
        """
        Clean up this joint
        """
        scene = self.scene()
        if scene:
            scene.removeItem(self._label)
            self._label = None
            scene.removeItem(self)
    # end def

    def _updateLabel(self, isLeft):
        """
        Called by updatePositionAndAppearance during init, or later by
        updateConnectivity. Updates drawing and position of the label.
        """
        lbl = self._label
        if self._idx != None:
            if lbl == None:
                bw = _baseWidth
                num = self._partnerVirtualHelix.number()
                tBR = _fm.tightBoundingRect(str(num))
                halfLabelH = tBR.height()/2.0
                halfLabelW = tBR.width()/2.0
                # determine x and y positions
                labelX = bw/2.0 - halfLabelW
                if self._isOnTop:
                    labelY = -0.25*halfLabelH - 0.5 - 0.5*bw
                else:
                    labelY = 2*halfLabelH + 0.5 + 0.5*bw
                # adjust x for left vs right
                labelXoffset = 0.25*bw if isLeft else -0.25*bw
                labelX += labelXoffset
                # adjust x for numeral 1
                if num == 1: labelX -= halfLabelW/2.0
                # create text item
                lbl = QGraphicsSimpleTextItem(str(num), self)
                lbl.setPos(labelX, labelY)
                lbl.setBrush(_enabbrush)
                lbl.setFont(_toHelixNumFont)
                self._label = lbl
            # end if
            lbl.setText( str(self._partnerVirtualHelix.number()) )
        # end if
    # end def

# end class


class XoverNode5(XoverNode3):
    """
    XoverNode5 is the partner of XoverNode3. It dif
    XoverNode3 handles:
    1. Drawing of the 5' end of an xover, and its text label. Drawing style
    is determined by the location of the xover with in a vhelix (is it a top
    or bottom vstrand?).
    2. Notifying XoverStrands in the model when connectivity changes.

    """
    def __init__(self, virtualHelixItem, xoverItem, strand5p, idx):
        super(XoverNode5, self).__init__(virtualHelixItem, xoverItem, strand5p, idx)
    # end def

    def setPartnerVirtualHelix(self, strand):
        if strand.connection3p():
            self._partnerVirtualHelix = strand.connection3p().virtualHelix()
        else:
            self._partnerVirtualHelix = None
    # end def

    def updatePositionAndAppearance(self):
        """Same as XoverItem3, but exposes 3' end"""
        self.setPos(*self.point())
        # # We can only expose a 3' end. But on which side?
        isLeft = False if self._isDrawn5to3 else True
        self._updateLabel(isLeft)
    # end def
# end class


class XoverItem(QGraphicsPathItem):
    """
    This class handles:
    1. Drawing the spline between the XoverNode3 and XoverNode5 graphics
    items in the path view.

    XoverItem should be a child of a PartItem.
    """
    _filterName = "xover"

    def __init__(self, strandItem, virtualHelixItem):
        """
        strandItem is a the model representation of the 5prime most strand
        of a Xover
        """
        super(XoverItem, self).__init__(virtualHelixItem.partItem())
        self._strandItem = strandItem
        self._virtualHelixItem = virtualHelixItem
        self._strand5p = None
        self._node5 = None
        self._node3 = None
        self.hide()
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)

        # for easier mouseclick
        self._clickArea = cA = QGraphicsRectItem(self)
        # self._clickArea.setAcceptHoverEvents(True)
        # cA.hoverMoveEvent = self.hoverMoveEvent
        cA.mousePressEvent = self.mousePressEvent
        cA.mouseMoveEvent = self.mouseMoveEvent
        cA.setPen(QPen(Qt.PenStyle.NoPen))
    # end def

    ### SLOTS ###

    ### ACCESSORS ###
    def activeTool(self):
        return self._strandItem._activeTool()
    # end def

    def partItem(self):
        return self._virtualHelixItem.partItem()
    # end def

    def remove(self):
        scene = self.scene()
        if self._node3:
            self._node3.remove()
            self._node5.remove()
            self._node3 = None
            self._node5 = None
        self._strand5p = None
        scene.removeItem(self._clickArea)
        self._clickArea = None
        scene.removeItem(self)
    # end def

    ### PUBLIC SUPPORT METHODS ###
    def hideIt(self):
        self.hide()
        if self._node3:
            self._node3.hide()
            self._node5.hide()
            self._node3.remove()
            self._node3 = None
    # end def

    def showIt(self):
        self.show()
        if self._node3:
            self._node3.show()
            self._node5.show()
    # end def

    def refreshXover(self):
        strand5p = self._strand5p
        node3 = self._node3
        if strand5p:
            strand3p = strand5p.connection3p()
            if strand3p != None and node3:
                if node3.virtualHelix():
                    self.update(self._strand5p)
                else:
                    node3.remove()
                    self._node3 = None
            elif node3:
                node3.remove()
                self._node3 = None
        elif node3:
            node3.remove()
            self._node3 = None
    # end def

    def update(self, strand5p, idx=None):
        """
        Pass idx to this method in order to install a floating
        Xover for the forced xover tool
        """
        self._strand5p = strand5p
        strand3p = strand5p.connection3p()
        vhi5p = self._virtualHelixItem
        partItem = vhi5p.partItem()

        # This condition is for floating xovers
        idx3Prime = idx if idx else strand5p.idx3Prime()

        if self._node5 == None:
            self._node5 = XoverNode5(vhi5p, self, strand5p, idx3Prime)
        if strand3p != None:
            if self._node3 == None:
                vhi3p = partItem.itemForVirtualHelix(strand3p.virtualHelix())
                self._node3 = XoverNode3(vhi3p, self, strand3p, strand3p.idx5Prime())
            else:
                self._node5.setIdx(idx3Prime)
                self._node3.setIdx(strand3p.idx5Prime())
            self._node5.setPartnerVirtualHelix(strand5p)
            self._updatePath(strand5p)
        else:
            if self._node3:
                self._node3.remove()
                self._node3 = None
        # end if
    # end def

    ### PRIVATE SUPPORT METHODS ###
    def _updatePath(self, strand5p):
        """
        Draws a quad curve from the edge of the fromBase
        to the top or bottom of the toBase (q5), and
        finally to the center of the toBase (toBaseEndpoint).

        If floatPos!=None, this is a floatingXover and floatPos is the
        destination point (where the mouse is) while toHelix, toIndex
        are potentially None and represent the base at floatPos.

        """
        group = self.group()
        self.tempReparent()

        node3 = self._node3
        node5 = self._node5

        bw = _baseWidth

        parent = self.partItem()

        vhi5 = self._virtualHelixItem
        pt5 = vhi5.mapToItem(parent, *node5.point())

        fiveIsTop = node5.isOnTop()
        fiveIs5to3 = node5.isDrawn5to3()

        vhi3 = node3.virtualHelixItem()
        pt3 = vhi3.mapToItem(parent, *node3.point())

        threeIsTop = node3.isOnTop()
        threeIs5to3 = node3.isDrawn5to3()
        sameStrand = (node5.strandType() == node3.strandType()) and vhi3 == vhi5
        sameParity = fiveIs5to3 == threeIs5to3

        # Enter/exit are relative to the direction that the path travels
        # overall.
        fiveEnterPt = pt5 + QPointF(0 if fiveIs5to3 else 1, .5)*bw
        fiveCenterPt = pt5 + QPointF(.5, .5)*bw
        fiveExitPt = pt5 + QPointF(.5, 0 if fiveIsTop else 1)*bw

        threeEnterPt = pt3 + QPointF(.5, 0 if threeIsTop else 1)*bw
        threeCenterPt = pt3 + QPointF(.5, .5)*bw
        threeExitPt = pt3 + QPointF(1 if threeIs5to3 else 0, .5)*bw

        c1 = QPointF()
        # case 1: same strand
        if sameStrand:
            dx = abs(threeEnterPt.x() - fiveExitPt.x())
            c1.setX(0.5 * (fiveExitPt.x() + threeEnterPt.x()))
            if fiveIsTop:
                c1.setY(fiveExitPt.y() - 0.05 * _yScale * dx)  # almost flat
            else:
                c1.setY(fiveExitPt.y() + 0.05 * _yScale * dx)  # almost flat
        # case 2: same parity
        elif sameParity:
            dy = abs(threeEnterPt.y() - fiveExitPt.y())
            c1.setX(fiveExitPt.x() + _xScale * dy)
            c1.setY(0.5 * (fiveExitPt.y() + threeEnterPt.y()))
        # case 3: different parity
        else:
            if fiveIsTop and fiveIs5to3:
                c1.setX(fiveExitPt.x() - _xScale *\
                        abs(threeEnterPt.y() - fiveExitPt.y()))
            else:
                c1.setX(fiveExitPt.x() + _xScale *\
                        abs(threeEnterPt.y() - fiveExitPt.y()))
            c1.setY(0.5 * (fiveExitPt.y() + threeEnterPt.y()))

        # Construct painter path
        painterpath = QPainterPath()
        painterpath.moveTo(fiveEnterPt)
        painterpath.lineTo(fiveCenterPt)
        painterpath.lineTo(fiveExitPt)

        # The xover5's non-crossing-over end (3') has a connection
        painterpath.quadTo(c1, threeEnterPt)
        painterpath.lineTo(threeCenterPt)
        painterpath.lineTo(threeExitPt)

        tempR = painterpath.boundingRect()
        tempR.adjust(-bw/2, 0, bw, 0)
        self._clickArea.setRect(tempR)
        self.setPath(painterpath)
        node3.updatePositionAndAppearance()
        node5.updatePositionAndAppearance()

        if group:
            group.addToGroup(self)

        self._updateColor(strand5p)
    # end def

    def updateLabels(self):
        if self._node3:
            self._node3._updateLabel()
        if self._node5:
            self._node5._updateLabel()

    def _updateColor(self, strand):
        oligo = strand.oligo()
        color = self.pen().color() if self.isSelected() else QColor(oligo.color())
        # print "update xover color", color.value(), self.isSelected(), self.group(), self.parentItem()
        penWidth = styles.PATH_STRAND_STROKE_WIDTH
        if oligo.shouldHighlight():
            penWidth = styles.PATH_STRAND_HIGHLIGHT_STROKE_WIDTH
            color.setAlpha(128)

        pen = QPen(color, penWidth)

        if self._node3 and self._node5:
            vhi5 = self._node5.virtualHelixItem()
            vhi3 = self._node3.virtualHelixItem()
            sameStrand = (self._node5.strandType() == self._node3.strandType()) and vhi3 == vhi5
            if sameStrand:
                pen.setStyle(Qt.PenStyle.DashLine)
                pen.setDashPattern([3, 2])

        pen.setCapStyle(Qt.PenCapStyle.FlatCap)
        self.setPen(pen)
    # end def

    ### EVENT HANDERS ###
    def mousePressEvent(self, event):
        """
        Special case for xovers and select tool, for now
        """
        if str(self.activeTool()) == "selectTool":
            event.setAccepted(False)
            sI = self._strandItem
            viewroot = sI.viewroot()
            currentFilterDict = viewroot.selectionFilterDict()
            if sI.strandFilter() in currentFilterDict and self._filterName in currentFilterDict:
                event.setAccepted(True)
                selectionGroup = viewroot.strandItemSelectionGroup()
                mod = Qt.KeyboardModifier.MetaModifier
                if not (event.modifiers() & mod):
                    selectionGroup.clearSelection(False)
                selectionGroup.setSelectionLock(selectionGroup)
                # self.setSelectedColor(True)
                selectionGroup.pendToAdd(self)
                selectionGroup.processPendingToAddList()
                return selectionGroup.mousePressEvent(event)
        else:
            event.setAccepted(False)
    # end def

    def eraseToolMousePress(self):
        """Erase the strand."""
        self._strandItem.eraseToolMousePress(None, None)
    # end def

    def paintToolMousePress(self):
        """Paint the strand."""
        self._strandItem.paintToolMousePress(None, None)
    # end def

    def selectToolMousePress(self):
        """Remove the xover."""
        # make sure the selection is clear
        sI = self._strandItem
        viewroot = sI.viewroot()
        selectionGroup = viewroot.strandItemSelectionGroup()
        selectionGroup.clearSelection(False)

        strand5p = self._strand5p
        strand3p = strand5p.connection3p()
        self._virtualHelixItem.part().removeXover(strand5p, strand3p)
    # end def

    def restoreParent(self, pos=None):
        """
        Required to restore parenting and positioning in the partItem
        """
        # map the position
        self.tempReparent(pos)
        self.setSelectedColor(False)
        self.setSelected(False)
    # end def

    def tempReparent(self, pos=None):
        partItem = self.partItem()
        if pos == None:
            pos = self.scenePos()
        self.setParentItem(partItem)
        tempP = partItem.mapFromScene(pos)
        self.setPos(tempP)
    # end def

    def setSelectedColor(self, value):
        if value == True:
            color = styles.selected_color
        else:
            oligo = self._strandItem.strand().oligo()
            color = QColor(oligo.color())
            if oligo.shouldHighlight():
                color.setAlpha(128)
        pen = self.pen()
        pen.setColor(color)
        self.setPen(pen)
    # end def

    def itemChange(self, change, value):
        # for selection changes test against QGraphicsItem.GraphicsItemChange.ItemSelectedChange
        # intercept the change instead of the has changed to enable features.
        if change == QGraphicsItem.GraphicsItemChange.ItemSelectedChange and self.scene():
            activeTool = self.activeTool()
            if str(activeTool) == "selectTool":
                sI = self._strandItem
                viewroot = sI.viewroot()
                currentFilterDict = viewroot.selectionFilterDict()
                selectionGroup = viewroot.strandItemSelectionGroup()
                # only add if the selectionGroup is not locked out
                if value == True and (self._filterName in currentFilterDict or not selectionGroup.isNormalSelect()):
                    if sI.strandFilter() in currentFilterDict:
                        # print "might add a xoi"
                        if self.group() != selectionGroup and selectionGroup.isNormalSelect():
                            # print "adding an xoi"
                            selectionGroup.pendToAdd(self)
                            selectionGroup.setSelectionLock(selectionGroup)
                        self.setSelectedColor(True)
                        return True
                    else:
                        # print "Doh"
                        return False
                # end if
                elif value == True:
                    # print "DOink"
                    return False
                else:
                    # Deselect
                    # Check if the strand is being added to the selection group still
                    if not selectionGroup.isPending(self._strandItem):
                        selectionGroup.pendToRemove(self)
                        self.tempReparent()
                        self.setSelectedColor(False)
                        return False
                    else:   # don't deselect it, because the strand is selected still
                        return True
                # end else
            # end if
            elif str(activeTool) == "paintTool":
                sI = self._strandItem
                viewroot = sI.viewroot()
                currentFilterDict = viewroot.selectionFilterDict()
                if sI.strandFilter() in currentFilterDict:
                    if not activeTool.isMacrod():
                        activeTool.setMacrod()
                    self.paintToolMousePress()
            return False
        # end if
        return QGraphicsPathItem.itemChange(self, change, value)
    # end def

    def modelDeselect(self, document):
        strand5p = self._strand5p
        strand3p = strand5p.connection3p()
        test5p = document.isModelStrandSelected(strand5p)
        lowVal5p, highVal5p = document.getSelectedStrandValue(strand5p) if test5p else (False, False)
        if strand5p.isDrawn5to3():
            highVal5p = False
        else:
            lowVal5p = False
        test3p = document.isModelStrandSelected(strand3p)
        lowVal3p, highVal3p = document.getSelectedStrandValue(strand3p) if test3p else (False, False)
        if strand3p.isDrawn5to3():
            lowVal3p = False
        else:
            highVal3p = False

        if not lowVal5p and not highVal5p and test5p:
            document.removeStrandFromSelection(strand5p)
        elif test5p:
            document.addStrandToSelection(strand5p, (lowVal5p, highVal5p))
        if not lowVal3p and not highVal3p and test3p:
            document.removeStrandFromSelection(strand3p)
        elif test3p:
            document.addStrandToSelection(strand3p, (lowVal3p, highVal3p))
        self.restoreParent()
    # end def

    def modelSelect(self, document):
        strand5p = self._strand5p
        strand3p = strand5p.connection3p()

        test5p = document.isModelStrandSelected(strand5p)
        lowVal5p, highVal5p = document.getSelectedStrandValue(strand5p) if test5p else (False, False)
        if strand5p.isDrawn5to3():
            highVal5p = True
        else:
            lowVal5p = True
        test3p = document.isModelStrandSelected(strand3p)
        lowVal3p, highVal3p = document.getSelectedStrandValue(strand3p) if test3p else (False, False)
        if strand3p.isDrawn5to3():
            lowVal3p = True
        else:
            highVal3p = True
        self.setSelectedColor(True)
        self.setSelected(True)
        document.addStrandToSelection(strand5p, (lowVal5p, highVal5p))
        document.addStrandToSelection(strand3p, (lowVal3p, highVal3p))
    # end def

    def paint(self, painter, option, widget):
        painter.setPen(self.pen())
        painter.setBrush(self.brush())
        painter.drawPath(self.path())
    # end def
# end class XoverItem



# Copyright 2011 Autodesk, Inc.  All rights reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# http://www.opensource.org/licenses/mit-license.php

"""
mayaSelectionContex.py
Created by Simon Breslav on 2011-09-27
Maya Selection Context, generates callbacks when things are selected
in Maya 3D view
"""

import sys
import maya.OpenMaya as OpenMaya
import maya.OpenMayaUI as OpenMayaUI
import maya.OpenMayaMPx as OpenMayaMPx
import maya.cmds as cmds
from controllers.mayacontrollers.mayaObjectManager import Mom

contextCmdName = "spMayaCtxCmd"


def selectionCallback(clientData):
    """
    Callback function that is called when the selection changes in Maya.
    """
    clientData.deleteManipulators()
    #print "mayaSelcectionContex: selectionCallback called"
    selectionList = OpenMaya.MSelectionList()
    OpenMaya.MGlobal.getActiveSelectionList(selectionList)
    selectionIter = OpenMaya.MItSelectionList(selectionList)
    decoratorList = []
    helixList = []

    m = Mom()

    manipulator = None
    manipObject = OpenMaya.MObject()

    while not selectionIter.isDone():
        transformNode = OpenMaya.MObject()
        dagNode = OpenMaya.MFnDagNode()
        try:
            selectionIter.getDependNode(transformNode)
            if transformNode.isNull() or not transformNode.hasFn( \
                                    OpenMaya.MFn.kDependencyNode):
                next(selectionIter)
                continue
            dagNode = OpenMaya.MFnDagNode(transformNode)
        except:
            next(selectionIter)
            continue
        if dagNode.name().startswith(m.decoratorTransformName):
            if dagNode.name() not in decoratorList:
                decoratorList.append(dagNode.name())
        elif dagNode.name().startswith(m.helixTransformName):
            Unused, HNumber = dagNode.name().split("_")
            helixName = "%s%s" % (m.helixNodeName, HNumber)
            helixNode = m.getNodeFromName(helixName)
            if helixNode:
                helixList.append(helixName)
                if manipulator is None:
                    manipulator = \
                            OpenMayaMPx.MPxManipContainer.newManipulator(
                                                "spHelixManip", manipObject)
                    if manipulator is not None:
                        clientData.addManipulator(manipObject)
                manipulator.addHelix(HNumber)
                #print "selectionCallback ", dagNode.name(), helixNode
        next(selectionIter)
    m.staplePreDecoratorSelected(decoratorList)
    if manipulator is not None:
        manipulator.finishedAddingHelices()
    m.strandsSelected(helixList)

    m.updateSelectionBoxes()


class mayaSelectionContext(OpenMayaMPx.MPxSelectionContext):
    """
    Maya Command that adds a callback function when something is selected
    """
    def __init__(self):
        """Initialize the commands"""
        OpenMayaMPx.MPxSelectionContext.__init__(self)

    def toolOnSetup(self, event):
        """Add the callback"""
        OpenMaya.MModelMessage.addCallback( \
                        OpenMaya.MModelMessage.kActiveListModified,
                                            selectionCallback, self)


class mayaSelctionCtxCmd(OpenMayaMPx.MPxContextCommand):
    def __init__(self):
        OpenMayaMPx.MPxContextCommand.__init__(self)

    def makeObj(self):
        return OpenMayaMPx.asMPxPtr(mayaSelectionContext())


def contextCmdCreator():
    """Wrapper function that created the command"""
    return OpenMayaMPx.asMPxPtr(mayaSelctionCtxCmd())


# initialize the script plug-in
def initializePlugin(mobject):
    """Register the context command with Maya"""
    mplugin = OpenMayaMPx.MFnPlugin(mobject)
    try:
        mplugin.registerContextCommand(contextCmdName, contextCmdCreator)
    except:
        print("Failed to register context command: %s" % contextCmdName)
        raise


# uninitialize the script plug-in
def uninitializePlugin(mobject):
    """Remove the context command"""
    mplugin = OpenMayaMPx.MFnPlugin(mobject)
    try:
        mplugin.deregisterContextCommand(contextCmdName)
    except:
        print("Failed to deregister context command: %s" % contextCmdName)
        raise



"""
helixitem.py
Created by Nick on 2011-09-28.
"""
import re
from cadnano2.cadnano import app
from cadnano2.model.enum import StrandType
from cadnano2.views import styles
import cadnano2.util as util

try:
    from OpenGL import GL
except:
    GL = False

GL = False

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['QPointF', 'QRectF', 'Qt'])
util.qtWrapImport('QtGui', globals(), ['QBrush', 'QFont', 'QPen', 'QDrag',
                                       'QTransform',
                                       'QUndoCommand'])
util.qtWrapImport('QtWidgets', globals(), ['QGraphicsPathItem',
                                           'QGraphicsEllipseItem',
                                           'QGraphicsItem',
                                           'QGraphicsRectItem',
                                           'QGraphicsSimpleTextItem',
                                           'QGraphicsTextItem',
                                           'QStyle'])

# strand addition stores some meta information in the UndoCommand's text
_strand_re = re.compile("\((\d+),(\d+)\)\.0\^(\d+)")


class EmptyHelixItem(QGraphicsEllipseItem):
    """docstring for EmptyHelixItem"""
    # set up default, hover, and active drawing styles
    _defaultBrush = QBrush(styles.grayfill)
    _defaultPen = QPen(styles.graystroke, styles.SLICE_HELIX_STROKE_WIDTH)
    _hoverBrush = QBrush(styles.bluefill)
    _hoverPen = QPen(styles.bluestroke, styles.SLICE_HELIX_HILIGHT_WIDTH)
    _radius = styles.SLICE_HELIX_RADIUS
    temp = styles.SLICE_HELIX_STROKE_WIDTH
    _defaultRect = QRectF(0, 0, 2 * _radius, 2 * _radius)
    temp = (styles.SLICE_HELIX_HILIGHT_WIDTH - temp)/2
    _hoverRect = _defaultRect.adjusted(-temp, -temp, temp, temp)
    _ZDefault = styles.ZSLICEHELIX 
    _ZHovered = _ZDefault+1 
    temp /= 2
    _adjustmentPlus = (temp, temp)
    _adjustmentMinus = (-temp, -temp)
    # _PI = 3.141592
    # _temp = [x*_PI*0.1 for x in range(20)]
    # _temp = [(math.sin(angle) * _radius, math.cos(angle) * _radius) for angle in _temp]

    def __init__(self, row, column, partItem):
        """
        row, column is a coordinate in Lattice terms
        partItem is a PartItem that will act as a QGraphicsItem parent
        """
        super(EmptyHelixItem, self).__init__(parent=partItem)
        self._partItem = partItem
        self._lastvh = None  # for decideAction
        self.hide()
        self._isHovered = False
        self.setAcceptHoverEvents(True)

        self.setNotHovered()

        x, y = partItem.part().latticeCoordToPositionXY(row, column, partItem.scaleFactor())
        self.setPos(x, y)
        self._coord = (row, column)
        self.show()
    # end def

    def virtualHelix(self):
        """
        virtualHelixItem should be the HelixItems only child if it exists
        and virtualHelix should be it member
        """
        temp = self.virtualHelixItem()
        if temp:
            return temp.virtualHelix()
        else:
            return None
    # end def

    def virtualHelixItem(self):
        """
        virtualHelixItem should be the HelixItems only child if it exists
        and virtualHelix should be it member
        """
        temp = self.childItems()
        if len(temp) > 0:
            return temp[0]
        else:
            return None
    # end def

    def part(self):
        return self._partItem.part()
    # end def

    def translateVH(self, delta):
        """
        used to update a child virtual helix position on a hover event
        delta is a tuple of x and y values to translate

        positive delta happens when hover happens
        negative delta when something unhovers
        """
        temp = self.virtualHelixItem()

        # xor the check to translate, 
        # convert to a QRectF adjustment if necessary
        check = (delta > 0) ^ self._isHovered
        if temp and check:
            pass
            # temp.translate(*delta)
    # end def

    def setHovered(self):
        # self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemHasNoContents, False)
        self.setBrush(self._hoverBrush)
        self.setPen(self._hoverPen)
        self.update(self.boundingRect())
        # self.translateVH(self._adjustmentPlus)
        self._isHovered = True
        self.setZValue(self._ZHovered)
        self.setRect(self._hoverRect)

        self._partItem.updateStatusBar("(%d, %d)" % self._coord)
    # end def

    def hoverEnterEvent(self, event):
        """
        hoverEnterEvent changes the HelixItem brush and pen from default
        to the hover colors if necessary.
        """
        self.setHovered()
    # end def

    def setNotHovered(self):
        """
        """
        # drawMe = False if self.virtualHelixItem() else True
        # self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemHasNoContents, drawMe)
        self.setBrush(self._defaultBrush)
        self.setPen(self._defaultPen)
        # self.translateVH(self._adjustmentMinus)
        self._isHovered = False
        self.setZValue(self._ZDefault)
        self.setRect(self._defaultRect)

        self._partItem.updateStatusBar("")
    # end def

    def hoverLeaveEvent(self, event):
        """
        hoverEnterEvent changes the HelixItem brush and pen from hover
        to the default colors if necessary.
        """
        self.setNotHovered()
    # end def

    def mousePressEvent(self, event):
        action = self.decideAction(event.modifiers())
        action(self)
        self.dragSessionAction = action
    # end def

    def mouseMoveEvent(self, event):
        partItem = self._partItem
        posInParent = partItem.mapFromItem(self, QPointF(event.pos()))
        # Qt doesn't have any way to ask for graphicsitem(s) at a
        # particular position but it *can* do intersections, so we
        # just use those instead
        partItem.probe.setPos(posInParent)
        for ci in partItem.probe.collidingItems():
            if isinstance(ci, EmptyHelixItem):
                self.dragSessionAction(ci)
    # end def

    def autoScafMidSeam(self, strands):
        """docstring for autoScafMidSeam"""
        part = self.part()
        strandType = StrandType.Scaffold
        idx = part.activeBaseIndex()
        for i in range(1, len(strands)):
            row1, col1, sSidx1 = strands[i-1]  # previous strand
            row2, col2, sSidx2 = strands[i]  # current strand
            vh1 = part.virtualHelixAtCoord((row1, col1))
            vh2 = part.virtualHelixAtCoord((row2, col2))
            strand1 = vh1.scaffoldStrandSet()._strandList[sSidx1]
            strand2 = vh2.scaffoldStrandSet()._strandList[sSidx2]
            # determine if the pair of strands are neighbors
            neighbors = part.getVirtualHelixNeighbors(vh1)
            if vh2 in neighbors:
                p2 = neighbors.index(vh2)
                if vh2.number() % 2 == 1:
                    # resize and install external xovers
                    try:
                        # resize to the nearest prexover on either side of idx
                        newLo = util.nearest(idx, part.getPreXoversHigh(strandType, p2, maxIdx=idx-10))
                        newHi = util.nearest(idx, part.getPreXoversLow(strandType, p2, minIdx=idx+10))
                        if strand1.canResizeTo(newLo, newHi) and \
                           strand2.canResizeTo(newLo, newHi):
                            # do the resize
                            strand1.resize((newLo, newHi))
                            strand2.resize((newLo, newHi))
                            # install xovers
                            part.createXover(strand1, newHi, strand2, newHi)
                            part.createXover(strand2, newLo, strand1, newLo)
                    except ValueError:
                        pass  # nearest not found in the expanded list

                    # go back an install the internal xovers
                    if i > 2:
                        row0, col0, sSidx0 = strands[i-2]  # two strands back
                        vh0 = part.virtualHelixAtCoord((row0, col0))
                        strand0 = vh0.scaffoldStrandSet()._strandList[sSidx0]
                        if vh0 in neighbors:
                            p0 = neighbors.index(vh0)
                            l0, h0 = strand0.idxs()
                            l1, h1 = strand1.idxs()
                            oLow, oHigh = util.overlap(l0, h0, l1, h1)
                            try:
                                lList = [x for x in part.getPreXoversLow(strandType, p0) if x>oLow and x<oHigh]
                                lX = lList[len(lList)//2]
                                hList = [x for x in part.getPreXoversHigh(strandType, p0) if x>oLow and x<oHigh]
                                hX = hList[len(hList)//2]
                                # install high xover first
                                part.createXover(strand0, hX, strand1, hX)
                                # install low xover after getting new strands
                                # following the breaks caused by the high xover
                                strand3 = vh0.scaffoldStrandSet()._strandList[sSidx0]
                                strand4 = vh1.scaffoldStrandSet()._strandList[sSidx1]
                                part.createXover(strand4, lX, strand3, lX)
                            except IndexError:
                                pass  # filter was unhappy

    def autoScafRaster(self, strands):
        """docstring for autoScafRaster"""
        part = self.part()
        idx = part.activeBaseIndex()
        for i in range(1, len(strands)):
            row1, col1, sSidx1 = strands[i-1]  # previous strand
            row2, col2, sSidx2 = strands[i]  # current strand
            vh1 = part.virtualHelixAtCoord((row1, col1))
            vh2 = part.virtualHelixAtCoord((row2, col2))
            strand1 = vh1.scaffoldStrandSet()._strandList[sSidx1]
            strand2 = vh2.scaffoldStrandSet()._strandList[sSidx2]
            # determine if the pair of strands are neighbors
            neighbors = part.getVirtualHelixNeighbors(vh1)
            if vh2 in neighbors:
                p2 = neighbors.index(vh2)
                if vh2.number() % 2 == 1:
                    # resize and install external xovers
                    try:
                        # resize to the nearest prexover on either side of idx
                        newLo1 = newLo2 = util.nearest(idx, part.getPreXoversHigh(StrandType.Scaffold, p2, maxIdx=idx-8))
                        newHi = util.nearest(idx, part.getPreXoversLow(StrandType.Scaffold, p2, minIdx=idx+8))

                        if vh1.number() != 0:  # after the first helix
                            newLo1 = strand1.lowIdx()  # leave alone the lowIdx

                        if vh2.number() != len(strands)-1:  # before the last
                            newLo2 = strand2.lowIdx()  # leave alone the lowIdx

                        if strand1.canResizeTo(newLo1, newHi) and \
                           strand2.canResizeTo(newLo2, newHi):
                            strand1.resize((newLo1, newHi))
                            strand2.resize((newLo2, newHi))
                        else:
                            raise ValueError
                        # install xovers
                        part.createXover(strand1, newHi, strand2, newHi)
                    except ValueError:
                        pass  # nearest not found in the expanded list
                else:
                    # resize and install external xovers
                    idx = part.activeBaseIndex()
                    try:
                        # resize to the nearest prexover on either side of idx
                        newLo = util.nearest(idx, part.getPreXoversHigh(StrandType.Scaffold, p2, maxIdx=idx-8))

                        if strand1.canResizeTo(newLo, strand1.highIdx()) and \
                           strand2.canResizeTo(newLo, strand2.highIdx()):
                            strand1.resize((newLo, strand1.highIdx()))
                            strand2.resize((newLo, strand2.highIdx()))
                            # install xovers
                            part.createXover(strand1, newLo, strand2, newLo)
                        else:
                            raise ValueError
                    except ValueError:
                        pass  # nearest not found in the expanded list

    def mouseReleaseEvent(self, event):
        """docstring for mouseReleaseEvent"""
        part = self.part()
        uS = part.undoStack()
        strands = []
        # Look at the undo stack in reverse order
        for i in range(uS.index()-1, 0, -1):
            # Check for contiguous strand additions
            m = _strand_re.match(uS.text(i))
            if m:
                strands.insert(0, list(map(int, m.groups())))
            else:
                break

        if len(strands) > 1:
            autoScafType = app().prefs.getAutoScafType()
            util.beginSuperMacro(part, "Auto-connect")
            if autoScafType == "Mid-seam":
                self.autoScafMidSeam(strands)
            elif autoScafType == "Raster":
                self.autoScafRaster(strands)
            util.endSuperMacro(part)

    def decideAction(self, modifiers):
        """ On mouse press, an action (add scaffold at the active slice, add
        segment at the active slice, or create virtualhelix if missing) is
        decided upon and will be applied to all other slices happened across by
        mouseMoveEvent. The action is returned from this method in the form of a
        callable function."""
        vh = self.virtualHelix()
        part = self.part()

        if vh == None:
            return EmptyHelixItem.addVHIfMissing

        idx = part.activeBaseIndex()
        scafSSet, stapSSet = vh.getStrandSets()
        if modifiers & Qt.KeyboardModifier.ShiftModifier:
            if not stapSSet.hasStrandAt(idx-1, idx+1):
                return EmptyHelixItem.addStapAtActiveSliceIfMissing
            else:
                return EmptyHelixItem.nop

        if not scafSSet.hasStrandAt(idx-1, idx+1):
            return EmptyHelixItem.addScafAtActiveSliceIfMissing
        return EmptyHelixItem.nop
    # end def

    def nop(self):
        self._partItem.updateStatusBar("(%d, %d)" % self._coord)

    def addScafAtActiveSliceIfMissing(self):
        vh = self.virtualHelix()
        part = self.part()
        if vh == None:
            return

        idx = part.activeBaseIndex()
        startIdx = max(0,idx-1)
        endIdx = min(idx+1, part.maxBaseIdx())
        vh.scaffoldStrandSet().createStrand(startIdx, endIdx)

        self._partItem.updateStatusBar("(%d, %d)" % self._coord)
    # end def

    def addStapAtActiveSliceIfMissing(self):
        vh = self.virtualHelix()
        part = self.part()

        if vh == None:
            return

        idx = part.activeBaseIndex()
        startIdx = max(0,idx-1)
        endIdx = min(idx+1, part.maxBaseIdx())
        vh.stapleStrandSet().createStrand(startIdx, endIdx)

        self._partItem.updateStatusBar("(%d, %d)" % self._coord)
    # end def

    def addVHIfMissing(self):
        vh = self.virtualHelix()
        coord = self._coord
        part = self.part()

        if vh != None:
            return
        uS = part.undoStack()
        uS.beginMacro("Slice Click")
        part.createVirtualHelix(*coord)
        # vh.scaffoldStrandSet().createStrand(startIdx, endIdx)
        uS.endMacro()

        self._partItem.updateStatusBar("(%d, %d)" % self._coord)
    # end def

    # if GL:
    #     def paint(self, painter, option, widget):
    #         painter.beginNativePainting()
    # 
    #         radius = self._radius
    # 
    #         # GL.glPushAttrib(GL.GL_ALL_ATTRIB_BITS)
    #         # GL.glClear(GL.GL_COLOR_BUFFER_BIT)
    # 
    #         # Draw the filled circle
    # 
    #         GL.glColor3f (1, 0.5, 0)       # Set to orange
    # 
    #         GL.glBegin (GL.GL_POLYGON)
    #         for X, Y in self._temp:
    #             GL.glVertex2f (X,Y)
    #         # end for
    #         GL.glEnd()
    # 
    #         # Draw the anti-aliased outline
    # 
    #         # GL.glEnable(GL.GL_BLEND)
    #         # GL.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA)
    #         # GL.glEnable(GL.GL_LINE_SMOOTH)
    # 
    #         # GL.glBegin(GL.GL_LINE_LOOP)
    #         # for angle in [x*PI*0.01 for x in range(200)]:
    #         #     GL.glVertex2f(X + math.sin(angle) * radius, Y + math.cos(angle) * radius)
    #         # # end for
    #         # GL.glDisable(GL.GL_BLEND)
    #         # GL.glEnd()
    #         # GL.glPopAttrib()
    #         painter.endNativePainting()
    #     # end def
    # # end if
